from __future__ import annotations

import datetime
import re
import subprocess
import time
from pathlib import Path

from briefcase.commands import (
    BuildCommand,
    CreateCommand,
    OpenCommand,
    PackageCommand,
    PublishCommand,
    RunCommand,
    UpdateCommand,
)
from briefcase.config import AppConfig, parsed_version
from briefcase.console import ANSI_ESC_SEQ_RE_DEF
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.android_sdk import AndroidSDK
from briefcase.integrations.subprocess import SubprocessArgT


def safe_formal_name(name):
    """Converts the name into a safe name on Android.

    Certain characters (``/\\:<>"?*|``) can't be used as app names
    on Android; ``!`` causes problems with Android build tooling.
    Also ensure that trailing, leading, and consecutive whitespace
    caused by removing punctuation is collapsed.

    :param name: The candidate name
    :returns: The safe version of the name.
    """
    return re.sub(r"\s+", " ", re.sub(r'[!/\\:<>"\?\*\|]', "", name)).strip()


# Matches zero or more ANSI control chars wrapping the message for when
# the Android emulator is printing in color.
ANDROID_LOG_PREFIX_REGEX = re.compile(
    rf"(?:{ANSI_ESC_SEQ_RE_DEF})*[A-Z]/(?P<tag>.*?): (?P<content>.*?(?=\x1B|$))(?:{ANSI_ESC_SEQ_RE_DEF})*"
)


def android_log_clean_filter(line):
    """Filter an ADB log to extract the Python-generated message content.

    Any system or stub messages are ignored; all logging prefixes are stripped.
    Python code is identified as coming from the ``python.stdout``

    :param line: The raw line from the system log
    :returns: A tuple, containing (a) the log line, stripped of any system
        logging context, and (b) a boolean indicating if the message should be
        included for analysis purposes (i.e., it's Python content, not a system
        message).
    """
    match = ANDROID_LOG_PREFIX_REGEX.match(line)
    if match:
        groups = match.groupdict()
        include = groups["tag"] in {"python.stdout", "python.stderr"}
        return groups["content"], include

    return line, False


class GradleMixin:
    output_format = "gradle"
    platform = "android"
    platform_target_version = "0.3.15"

    @property
    def packaging_formats(self):
        return ["aab", "apk", "debug-apk"]

    @property
    def default_packaging_format(self):
        return "aab"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def project_path(self, app):
        return self.bundle_path(app)

    def binary_path(self, app):
        return (
            self.bundle_path(app)
            / "app"
            / "build"
            / "outputs"
            / "apk"
            / "debug"
            / "app-debug.apk"
        )

    def distribution_path(self, app):
        extension = {
            "aab": "aab",
            "apk": "apk",
            "debug-apk": "debug.apk",
        }[app.packaging_format]
        return self.dist_path / f"{app.formal_name}-{app.version}.{extension}"

    def run_gradle(self, app, args: list[SubprocessArgT]):
        # Gradle may install the emulator via the dependency chain build-tools > tools >
        # emulator. (The `tools` package only shows up in sdkmanager if you pass
        # `--include_obsolete`.) However, the old sdkmanager built into Android Gradle
        # plugin 4.2 doesn't know about macOS on ARM, so it'll install an x86_64 emulator
        # which won't work with ARM system images.
        #
        # Work around this by pre-installing the emulator with our own sdkmanager before
        # running Gradle. For simplicity, we do this on all platforms, since the user will
        # almost certainly want an emulator soon enough.
        self.tools.android_sdk.verify_emulator()

        gradlew = "gradlew.bat" if self.tools.host_os == "Windows" else "gradlew"
        self.tools.subprocess.run(
            # Windows needs the full path to `gradlew`; macOS & Linux can find it
            # via `./gradlew`. For simplicity of implementation, we always provide
            # the full path.
            [
                self.bundle_path(app) / gradlew,
                "--console",
                "plain",
            ]
            + (["--debug"] if self.tools.logger.is_deep_debug else [])
            + args,
            env=self.tools.android_sdk.env,
            # Set working directory so gradle can use the app bundle path as its
            # project root, i.e., to avoid 'Task assembleDebug not found'.
            cwd=self.bundle_path(app),
            check=True,
            # Gradle writes to stdout using the system encoding. So, explicitly use it
            # here to avoid defaulting to the console encoding for the subprocess call.
            # This is mostly for the benefit of Windows where the system encoding may
            # not be the same as the console encoding and typically neither are UTF-8.
            # See #1425 for details.
            encoding=self.tools.system_encoding,
        )

    def verify_tools(self):
        """Verify that the Android APK tools in `briefcase` will operate on this system,
        downloading tools as needed."""
        super().verify_tools()
        AndroidSDK.verify(tools=self.tools)
        if not self.is_clone:
            self.logger.add_log_file_extra(self.tools.android_sdk.list_packages)


class GradleCreateCommand(GradleMixin, CreateCommand):
    description = "Create and populate an Android Gradle project."
    hidden_app_properties = {"permission", "feature"}

    def support_package_filename(self, support_revision):
        """The query arguments to use in a support package query request."""
        return (
            f"Python-{self.python_version_tag}-Android-support.b{support_revision}.zip"
        )

    def output_format_template_context(self, app: AppConfig):
        """Additional template context required by the output format.

        :param app: The config object for the app
        """
        # Android requires an integer "version code". If a version code
        # isn't explicitly provided, generate one from the version number.
        # The build number will also be appended, if provided.
        try:
            version_code = app.version_code
        except AttributeError:
            parsed = parsed_version(app.version)

            v = (list(parsed.release) + [0, 0])[:3]  # version triple
            build = int(getattr(app, "build", "0"))
            version_code = f"{v[0]:d}{v[1]:02d}{v[2]:02d}{build:02d}".lstrip("0")

        # The default runtime libraries included in an app. The default value is the
        # list that was hard-coded in the Briefcase 0.3.16 Android template, prior to
        # the introduction of customizable system requirements for Android.
        try:
            dependencies = app.build_gradle_dependencies
        except AttributeError:
            self.logger.warning(
                """
*************************************************************************
** WARNING: App does not define build_gradle_dependencies              **
*************************************************************************

    The Android configuration for this app does not contain a
    `build_gradle_dependencies` definition. Briefcase will use a default
    value of:

        build_gradle_dependencies = [
            "androidx.appcompat:appcompat:1.0.2",
            "androidx.constraintlayout:constraintlayout:1.1.3",
            "androidx.swiperefreshlayout:swiperefreshlayout:1.1.0",
        ]

    You should add this definition to the Android configuration
    of your project's pyproject.toml file. See:

        https://briefcase.readthedocs.io/en/stable/reference/platforms/android.html#build_gradle-dependencies

    for more information.

*************************************************************************

"""
            )
            dependencies = [
                "androidx.appcompat:appcompat:1.0.2",
                "androidx.constraintlayout:constraintlayout:1.1.3",
                "androidx.swiperefreshlayout:swiperefreshlayout:1.1.0",
            ]

        return {
            "version_code": version_code,
            "safe_formal_name": safe_formal_name(app.formal_name),
            # Extract test packages, to enable features like test discovery and assertion
            # rewriting.
            "extract_packages": ", ".join(
                f'"{name}"'
                for path in (app.test_sources or [])
                if (name := Path(path).name)
            ),
            "build_gradle_dependencies": {"implementation": dependencies},
        }

    def permissions_context(self, app: AppConfig, x_permissions: dict[str, str]):
        """Additional template context for permissions.

        :param app: The config object for the app
        :param x_permissions: The dictionary of known cross-platform permission
            definitions.
        :returns: The template context describing permissions for the app.
        """
        # Default permissions for all Android apps
        permissions = {
            "android.permission.INTERNET": True,
            "android.permission.ACCESS_NETWORK_STATE": True,
        }

        # Default feature usage for all Android apps
        features = {}

        if x_permissions["camera"]:
            permissions["android.permission.CAMERA"] = True
            features["android.hardware.camera"] = False
            features["android.hardware.camera.any"] = False
            features["android.hardware.camera.front"] = False
            features["android.hardware.camera.external"] = False
            features["android.hardware.camera.autofocus"] = False

        if x_permissions["microphone"]:
            permissions["android.permission.RECORD_AUDIO"] = True

        if x_permissions["fine_location"]:
            permissions["android.permission.ACCESS_FINE_LOCATION"] = True
            features["android.hardware.location.network"] = False
            features["android.hardware.location.gps"] = False

        if x_permissions["coarse_location"]:
            permissions["android.permission.ACCESS_COARSE_LOCATION"] = True
            features["android.hardware.location.network"] = False
            features["android.hardware.location.gps"] = False

        if x_permissions["background_location"]:
            permissions["android.permission.ACCESS_BACKGROUND_LOCATION"] = True
            features["android.hardware.location.network"] = False
            features["android.hardware.location.gps"] = False

        if x_permissions["photo_library"]:
            permissions["android.permission.READ_MEDIA_VISUAL_USER_SELECTED"] = True

        # Override any permission and entitlement definitions with the platform specific definitions
        permissions.update(app.permission)
        features.update(getattr(app, "feature", {}))

        return {
            "permissions": permissions,
            "features": features,
        }


class GradleUpdateCommand(GradleCreateCommand, UpdateCommand):
    description = "Update an existing Android Gradle project."


class GradleOpenCommand(GradleMixin, OpenCommand):
    description = "Open the folder for an existing Android Gradle project."


class GradleBuildCommand(GradleMixin, BuildCommand):
    description = "Build an Android debug APK."

    def metadata_resource_path(self, app: AppConfig):
        return self.bundle_path(app) / self.path_index(app, "metadata_resource_path")

    def update_app_metadata(self, app: AppConfig, test_mode: bool):
        with self.input.wait_bar("Setting main module..."):
            with self.metadata_resource_path(app).open("w", encoding="utf-8") as f:
                # Set the name of the app's main module; this will depend
                # on whether we're in test mode.
                f.write(
                    f"""\
<resources>
    <string name="main_module">{app.main_module(test_mode)}</string>
</resources>
"""
                )

    def build_app(self, app: AppConfig, test_mode: bool, **kwargs):
        """Build an application.

        :param app: The application to build
        :param test_mode: Should the app be updated in test mode? (default: False)
        """
        self.logger.info("Updating app metadata...", prefix=app.app_name)
        self.update_app_metadata(app=app, test_mode=test_mode)

        self.logger.info("Building Android APK...", prefix=app.app_name)
        with self.input.wait_bar("Building..."):
            try:
                self.run_gradle(app, ["assembleDebug"])
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError("Error while building project.") from e


class GradleRunCommand(GradleMixin, RunCommand):
    description = "Run an Android debug APK on a device (physical or virtual)."

    def verify_tools(self):
        super().verify_tools()
        self.tools.android_sdk.verify_emulator()

    def add_options(self, parser):
        super().add_options(parser)
        parser.add_argument(
            "-d",
            "--device",
            dest="device_or_avd",
            help=(
                "The device to target; either a device ID for a physical device, "
                " or an AVD name ('@emulatorName') "
            ),
            required=False,
        )
        parser.add_argument(
            "--Xemulator",
            action="append",
            dest="extra_emulator_args",
            help="Additional arguments to use when starting the emulator",
            required=False,
        )
        parser.add_argument(
            "--shutdown-on-exit",
            action="store_true",
            help="Shutdown the emulator on exit",
            required=False,
        )

    def run_app(
        self,
        app: AppConfig,
        test_mode: bool,
        passthrough: list[str],
        device_or_avd=None,
        extra_emulator_args=None,
        shutdown_on_exit=False,
        **kwargs,
    ):
        """Start the application.

        :param app: The config object for the app
        :param test_mode: Boolean; Is the app running in test mode?
        :param passthrough: The list of arguments to pass to the app
        :param device_or_avd: The device to target. If ``None``, the user will
            be asked to re-run the command selecting a specific device.
        :param extra_emulator_args: Any additional arguments to pass to the emulator.
        :param shutdown_on_exit: Should the emulator be shut down on exit?
        """
        device, name, avd = self.tools.android_sdk.select_target_device(device_or_avd)

        # If there's no device ID, that means the emulator isn't running.
        # If there's no AVD either, it means the user has chosen to create
        # an entirely new emulator. Create the emulator (if necessary),
        # then start it.
        if device is None:
            if avd is None:
                avd = self.tools.android_sdk.create_emulator()
            else:
                # Ensure the system image for the requested emulator is available.
                # This step is only needed if the AVD already existed; you have to
                # have an image available to create an AVD.
                self.tools.android_sdk.verify_avd(avd)

            if extra_emulator_args:
                extra = f" (with {' '.join(extra_emulator_args)})"
            else:
                extra = ""
            self.logger.info(f"Starting emulator {avd}{extra}...", prefix=app.app_name)
            device, name = self.tools.android_sdk.start_emulator(
                avd, extra_emulator_args
            )

        try:
            label = "test suite" if test_mode else "app"

            self.logger.info(
                f"Starting {label} on {name} (device ID {device})", prefix=app.app_name
            )

            # Create an ADB wrapper for the selected device
            adb = self.tools.android_sdk.adb(device=device)

            # Compute Android package name. The Android template uses
            # `package_name` and `module_name`, so we use those here as well.
            package = f"{app.package_name}.{app.module_name}"

            # We force-stop the app to ensure the activity launches freshly.
            self.logger.info("Installing app...", prefix=app.app_name)
            with self.input.wait_bar("Stopping old versions of the app..."):
                adb.force_stop_app(package)

            # Install the latest APK file onto the device.
            with self.input.wait_bar("Installing new app version..."):
                adb.install_apk(self.binary_path(app))

            # To start the app, we launch `org.beeware.android.MainActivity`.
            with self.input.wait_bar(f"Launching {label}..."):
                # capture the earliest time for device logging in case PID not found
                device_start_time = adb.datetime()

                adb.start_app(package, "org.beeware.android.MainActivity", passthrough)

                # Try to get the PID for 5 seconds.
                pid = None
                fail_time = datetime.datetime.now() + datetime.timedelta(seconds=5)
                while not pid and datetime.datetime.now() < fail_time:
                    # Try to get the PID; run in quiet mode because we may
                    # need to do this a lot in the next 5 seconds.
                    pid = adb.pidof(package, quiet=True)
                    if not pid:
                        time.sleep(0.01)

            if pid:
                self.logger.info(
                    "Following device log output (type CTRL-C to stop log)...",
                    prefix=app.app_name,
                )
                # Start adb's logcat in a way that lets us stream the logs
                log_popen = adb.logcat(pid=pid)

                # Stream the app logs.
                self._stream_app_logs(
                    app,
                    popen=log_popen,
                    test_mode=test_mode,
                    clean_filter=android_log_clean_filter,
                    clean_output=False,
                    # Check for the PID in quiet mode so logs aren't corrupted.
                    stop_func=lambda: not adb.pid_exists(pid=pid, quiet=True),
                    log_stream=True,
                )
            else:
                self.logger.error("Unable to find PID for app", prefix=app.app_name)
                self.logger.error("Logs for launch attempt follow...")
                self.logger.error("=" * 75)

                # Show the log from the start time of the app
                adb.logcat_tail(since=device_start_time)

                raise BriefcaseCommandError(f"Problem starting app {app.app_name!r}")
        finally:
            if shutdown_on_exit:
                with self.tools.input.wait_bar("Stopping emulator..."):
                    adb.kill()


class GradlePackageCommand(GradleMixin, PackageCommand):
    description = "Create an Android App Bundle and APK in release mode."

    def package_app(self, app: AppConfig, **kwargs):
        """Package the app for distribution.

        This involves building the release app bundle.

        :param app: The application to build
        """
        self.logger.info(
            "Building Android App Bundle and APK in release mode...",
            prefix=app.app_name,
        )
        with self.input.wait_bar("Bundling..."):
            build_type, build_artefact_path = {
                "aab": ("bundleRelease", "bundle/release/app-release.aab"),
                "apk": ("assembleRelease", "apk/release/app-release-unsigned.apk"),
                "debug-apk": ("assembleDebug", "apk/debug/app-debug.apk"),
            }[app.packaging_format]

            try:
                self.run_gradle(app, [build_type])
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError("Error while building project.") from e

        # Move artefact to final location.
        self.tools.shutil.move(
            self.bundle_path(app) / "app/build/outputs" / build_artefact_path,
            self.distribution_path(app),
        )


class GradlePublishCommand(GradleMixin, PublishCommand):
    description = "Publish an Android APK."


# Declare the briefcase command bindings
create = GradleCreateCommand
open = GradleOpenCommand
update = GradleUpdateCommand
build = GradleBuildCommand
run = GradleRunCommand
package = GradlePackageCommand
publish = GradlePublishCommand



import subprocess

import pytest

from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.xcode import XcodeCliTools


def test_not_installed(mock_tools):
    """If cmdline dev tools are not installed, raise an error."""
    with pytest.raises(BriefcaseCommandError):
        XcodeCliTools.ensure_command_line_tools_are_installed(mock_tools)

    # xcode-select was invoked
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["xcode-select", "--install"],
    )


def test_installed(capsys, mock_tools):
    """If cmdline dev tools *are* installed, check passes without comment."""
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        cmd=["xcode-select", "--install"],
        returncode=1,
    )

    # Check passes without an error...
    XcodeCliTools.ensure_command_line_tools_are_installed(mock_tools)

    # ... xcode-select was invoked
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["xcode-select", "--install"],
    )

    # ...and the user is none the wiser
    out = capsys.readouterr().out
    assert len(out) == 0


def test_unsure_if_installed(capsys, mock_tools):
    """If xcode-select returns something odd, mention it but don't break."""
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        cmd=["xcode-select", "--install"],
        returncode=69,
    )

    # Check passes without an error...
    XcodeCliTools.ensure_command_line_tools_are_installed(mock_tools)

    # ... xcode-select was invoked
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["xcode-select", "--install"]
    )

    # ...but stdout contains a warning
    out = capsys.readouterr().out
    assert "************" in out



from __future__ import annotations

import subprocess
from unittest.mock import MagicMock, call

import pytest

import briefcase
from briefcase.config import AppConfig
from briefcase.integrations.base import ToolCache
from briefcase.integrations.docker import Docker, DockerAppContext


@pytest.fixture
def mock_tools(mock_tools, tmp_path) -> ToolCache:
    # Mock stdlib subprocess module
    mock_tools.subprocess._subprocess = MagicMock(spec_set=subprocess)

    # Reset `os` mock without `spec` so tests can run on Windows where os.getuid doesn't exist.
    mock_tools.os = MagicMock()
    # Mock user and group IDs for docker image
    mock_tools.os.getuid.return_value = "37"
    mock_tools.os.getgid.return_value = "42"

    # Mock return values for run()
    run_result = MagicMock(spec=subprocess.CompletedProcess, returncode=3)
    mock_tools.subprocess._subprocess.run.return_value = run_result

    # Mock return values for check_output()
    mock_tools.subprocess._subprocess.check_output.return_value = "goodbye\n"

    # Mock the return value for Popen()
    popen_process = MagicMock(spec_set=subprocess.Popen)
    mock_tools.subprocess._subprocess.Popen = MagicMock(return_value=popen_process)
    # preserve the Popen process for test assertions
    mock_tools._popen_process = popen_process

    # Mock the Wait Bar streamer
    wait_bar_streamer = MagicMock()
    wait_bar_streamer.stdout.readline.return_value = ""
    wait_bar_streamer.poll.return_value = 0
    # assume using the Popen process in a context manager is for Wait Bar
    popen_process.__enter__.return_value = wait_bar_streamer

    return mock_tools


@pytest.fixture
def my_app() -> AppConfig:
    return AppConfig(
        app_name="myapp",
        formal_name="My App",
        bundle="com.example",
        version="1.2.3",
        description="This is a simple app",
        license={"file": "LICENSE"},
        sources=["path/to/src/myapp", "other/stuff"],
        system_requires=["things==1.2", "stuff>=3.4"],
        system_runtime_requires=["runtime_things==1.42", "stuff>=3.4"],
    )


@pytest.fixture
def mock_docker(mock_tools, monkeypatch) -> Docker:
    """Adds a mocked Docker to the mock_tools."""
    # Default to mapped users to avoid subprocess calls
    monkeypatch.setattr(
        Docker, "_is_user_mapping_enabled", MagicMock(return_value=True)
    )

    mock_tools.docker = Docker(mock_tools)

    mock_tools.os.environ = {"PROCESS_ENV_VAR": "VALUE"}

    # Reset the mock so that the user mapping calls don't appear in test results
    mock_tools.subprocess._subprocess.check_output.reset_mock()

    # Reset the LRU cache for Docker.cache_image() for each test
    Docker.cache_image.cache_clear()

    return mock_tools.docker


@pytest.fixture
def mock_docker_app_context(mock_tools, my_app, tmp_path) -> DockerAppContext:
    """Adds a mocked DockerAppContext to the mock_tools."""
    mock_tools[my_app].app_context = DockerAppContext(mock_tools, my_app)
    mock_tools[my_app].app_context.prepare(
        image_tag="briefcase/com.example.myapp:py3.X",
        dockerfile_path=tmp_path / "bundle/Dockerfile",
        app_base_path=tmp_path / "base",
        host_bundle_path=tmp_path / "bundle",
        host_data_path=tmp_path / "briefcase",
        python_version="3.X",
    )

    # Allow for asserting _dockerize_args() calls but still run logic
    mock_tools[my_app].app_context._dockerize_args = MagicMock(
        wraps=mock_tools[my_app].app_context._dockerize_args
    )

    # Reset the mock so that the prepare call doesn't appear in test results
    mock_tools[my_app].app_context.tools.subprocess._subprocess.Popen.reset_mock()

    return mock_tools[my_app].app_context


@pytest.fixture
def user_mapping_run_calls(tmp_path, monkeypatch) -> list[call]:
    """The series of calls for determining how users are mapped."""
    monkeypatch.setattr(
        briefcase.integrations.docker.Docker,
        "_write_test_path",
        MagicMock(return_value=tmp_path / "build/mock_write_test"),
    )
    return [
        call(
            ["docker", "images", "-q", "alpine"],
            env={"DOCKER_CLI_HINTS": "false"},
        ),
        call(
            args=[
                "docker",
                "run",
                "--rm",
                "--volume",
                f"{tmp_path / 'build'}:/host_write_test:z",
                "alpine",
                "touch",
                "/host_write_test/mock_write_test",
            ],
            env={"DOCKER_CLI_HINTS": "false"},
        ),
        call(
            args=[
                "docker",
                "run",
                "--rm",
                "--volume",
                f"{tmp_path / 'build'}:/host_write_test:z",
                "alpine",
                "rm",
                "-f",
                "/host_write_test/mock_write_test",
            ],
            env={"DOCKER_CLI_HINTS": "false"},
        ),
    ]



import os
import subprocess
from pathlib import Path
from unittest import mock

import pytest

from briefcase.console import Console
from briefcase.exceptions import BriefcaseError
from briefcase.integrations.base import ToolCache
from briefcase.integrations.subprocess import Subprocess
from briefcase.integrations.xcode import get_simulators


@pytest.fixture
def mock_tools(mock_tools) -> ToolCache:
    mock_tools.input = mock.MagicMock(spec_set=Console)
    mock_tools.subprocess = Subprocess(mock_tools)
    mock_tools.subprocess._subprocess = mock.MagicMock(spec_set=subprocess)
    mock_tools.subprocess.check_output = mock.MagicMock()
    return mock_tools


@pytest.fixture
def simulator(tmp_path):
    """Create a dummy location for the simulator."""
    simulator_location = tmp_path / "CoreSimulator.framework"
    simulator_location.mkdir(parents=True, exist_ok=True)
    return os.fsdecode(simulator_location)


def simctl_result(name):
    """Load a simctl result file from the sample directory, and return the content."""
    filename = Path(__file__).parent / "simctl" / f"{name}.json"
    with filename.open(encoding="utf-8") as f:
        return f.read()


def test_simulator_is_missing(mock_tools, tmp_path):
    """If the simulator is not installed, a prompt is shown to the user."""
    mock_tools.subprocess.check_output.return_value = simctl_result("no-runtimes")

    simulators = get_simulators(
        mock_tools,
        "iOS",
        simulator_location=tmp_path / "CoreSimulator.framework",
    )

    # The prompt was displayed
    assert mock_tools.input.call_count == 1

    # The call returned without error.
    assert simulators == {}


def test_simctl_missing(mock_tools, simulator):
    """If simctl is missing or fails to start, an exception is raised."""
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        cmd=["xcrun", "simctl", "list", "-j"], returncode=1
    )

    with pytest.raises(BriefcaseError, match="Unable to run xcrun simctl."):
        get_simulators(mock_tools, "iOS", simulator_location=simulator)


def test_simctl_output_parse_error(mock_tools, simulator):
    """If parsing simctl JSON output fails, an exception is raised."""
    mock_tools.subprocess.check_output.return_value = "this is not JSON"

    with pytest.raises(BriefcaseError, match="Unable to parse output of xcrun simctl"):
        get_simulators(mock_tools, "iOS", simulator_location=simulator)


def test_no_runtimes(mock_tools, simulator):
    """If there are no runtimes available, no simulators will be found."""
    mock_tools.subprocess.check_output.return_value = simctl_result("no-runtimes")

    simulators = get_simulators(
        mock_tools,
        "iOS",
        simulator_location=simulator,
    )

    assert simulators == {}

    # The prompt was not shown
    mock_tools.input.assert_not_called()


def test_single_iOS_runtime(mock_tools, simulator):
    """If an iOS version is installed, devices can be found."""
    mock_tools.subprocess.check_output.return_value = simctl_result("iOS-13.2-only")

    simulators = get_simulators(
        mock_tools,
        "iOS",
        simulator_location=simulator,
    )

    assert simulators == {
        "iOS 13.2": {
            "20C5B052-F47A-4816-8584-9F1500B50477": "iPad Pro (9.7-inch)",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
            "314E772A-8034-44B4-9B28-3EE80C958F0A": "iPad Pro (12.9-inch) (3rd generation)",
            "36E4663B-A10F-470F-94E8-05C3DC692AC9": "iPad Pro (11-inch)",
            "5497F9B2-F4F3-454A-A9DD-993DF44EBB63": "iPhone 8 Plus",
            "939B1EF6-C25A-4056-B61F-20A2835E89D6": "iPad (7th generation)",
            "9CFF88F3-489F-444E-8131-FF8731768D31": "iPhone 11 Pro",
            "B490A004-6B5C-4C2C-BE4D-ACAB5D36C25D": "iPad Air (3rd generation)",
            "C9A005C8-9468-47C5-8376-68A6E3408209": "iPhone 8",
            "EEEBA06C-81F9-407C-885A-2261306DB2BE": "iPhone 11 Pro Max",
        }
    }

    # The prompt was not shown
    mock_tools.input.assert_not_called()


def test_watchOS_runtime(mock_tools, simulator):
    """Runtimes other than iOS can be requested."""
    mock_tools.subprocess.check_output.return_value = simctl_result("iOS-13.2-only")

    simulators = get_simulators(
        mock_tools,
        "watchOS",
        simulator_location=simulator,
    )

    assert simulators == {
        "watchOS 6.1": {
            "3EE6E657-9351-406C-9B39-24F0CECCBC74": "Apple Watch Series 5 - 40mm",
            "3EE83472-A457-4531-A221-67E332359EEC": "Apple Watch Series 4 - 40mm",
            "ABC5ABF6-C24E-4500-ADDC-A9375FFC36F6": "Apple Watch Series 5 - 44mm",
            "C240F238-3C85-46F1-AC40-D32C7896D430": "Apple Watch Series 4 - 44mm",
        }
    }

    # The prompt was not shown
    mock_tools.input.assert_not_called()


def test_multiple_iOS_runtime(mock_tools, simulator):
    """If multiple iOS versions are installed, this will be reflected in results."""
    mock_tools.subprocess.check_output.return_value = simctl_result(
        "multiple-iOS-versions"
    )

    simulators = get_simulators(
        mock_tools,
        "iOS",
        simulator_location=simulator,
    )

    assert simulators == {
        "iOS 13.2": {
            "20C5B052-F47A-4816-8584-9F1500B50477": "iPad Pro (9.7-inch)",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
            "314E772A-8034-44B4-9B28-3EE80C958F0A": "iPad Pro (12.9-inch) (3rd generation)",
            "36E4663B-A10F-470F-94E8-05C3DC692AC9": "iPad Pro (11-inch)",
            "5497F9B2-F4F3-454A-A9DD-993DF44EBB63": "iPhone 8 Plus",
            "939B1EF6-C25A-4056-B61F-20A2835E89D6": "iPad (7th generation)",
            "9CFF88F3-489F-444E-8131-FF8731768D31": "iPhone 11 Pro",
            "B490A004-6B5C-4C2C-BE4D-ACAB5D36C25D": "iPad Air (3rd generation)",
            "C9A005C8-9468-47C5-8376-68A6E3408209": "iPhone 8",
            "EEEBA06C-81F9-407C-885A-2261306DB2BE": "iPhone 11 Pro Max",
        },
        "iOS 10.3": {
            "04D415A8-FBF9-42AD-9F79-0CD452FA09D8": "iPad Pro (9.7 inch)",
            "0BB80120-FA02-4597-A1BA-DB8CDE4F086D": "iPhone 5s",
            "1E2A9CEF-AC9B-4CE3-AA66-9EBD785AAF23": "iPhone 6 Plus",
            "28D58EF9-C079-4809-A9E7-DA0BD7170DA6": "iPad Air 2",
            "425DE89A-49EA-44B0-9B2E-55BB18D4EF42": "iPhone SE",
            "6998CA09-44B5-4963-8F80-265412D99683": "iPhone 7",
            "6CF6870A-A7A6-4A42-8FE7-26FC3E131CDD": "iPad (5th generation)",
            "A2326477-DD25-4A61-A529-C772E364EDF3": "iPad Pro (10.5-inch)",
            "B7DEA42B-4740-4542-A1D5-8827546F531B": "iPhone 7 Plus",
            "BD9BF79E-92E2-444F-8708-4041B7D03681": "iPad Air",
            "CC954566-F315-4692-A754-DECDF72967CD": "iPhone 5",
            "D7BBAD14-38FD-48F5-ACFD-B1193F829216": "iPhone 6",
            "DA9B9F49-A070-4FD7-A6BF-8F49DC72194E": "iPhone 6s Plus",
            "E582FF8E-A5DC-4985-B6C8-8D6B1795DF62": "iPad Pro (12.9-inch) (2nd generation)",
            "E956D6AE-29F5-4780-A02A-D3426B7B4018": "iPad Pro (12.9 inch)",
            "F9A6C462-9A4A-438A-B541-848F0E6DBE5A": "iPhone 6s",
        },
    }

    # The prompt was not shown
    mock_tools.input.assert_not_called()


def test_unknown_runtime(mock_tools, simulator):
    """If an unknown runtime is requested, no devices will be found."""
    mock_tools.subprocess.check_output.return_value = simctl_result(
        "multiple-iOS-versions"
    )

    simulators = get_simulators(
        mock_tools,
        "whizzOS",
        simulator_location=simulator,
    )

    assert simulators == {}

    # The prompt was not shown
    mock_tools.input.assert_not_called()


def test_alternate_format(mock_tools, simulator):
    """The alternate format for device versions can be parsed."""
    mock_input = mock.MagicMock()

    mock_tools.subprocess.check_output.return_value = simctl_result("alternate-format")

    simulators = get_simulators(
        mock_tools,
        "iOS",
        simulator_location=simulator,
    )

    assert simulators == {
        "iOS 12.0": {
            "00751854-B15C-4C77-ACF7-2657CEF7DB58": "iPhone 7",
            "109F9913-9D7C-4170-974D-49DD6ED30AFE": "iPhone XR",
            "184397FC-2FBD-4342-9B93-C3B9710C86CD": "iPhone 6s Plus",
            "2C44F495-398B-4613-B798-D8DE8E3A1845": "iPhone XS",
            "354F7547-6A8B-4191-9A89-5EB707D6A1F0": "iPhone 8",
            "3A6C95FA-E3E9-4698-8E51-7EC60C4A642E": "iPad Pro (10.5-inch)",
            "485AA24A-BABF-4635-B849-8EE7BA10F259": "iPad (5th generation)",
            "497C2D40-6AA3-4485-B9DA-0C9615511854": "iPad Air 2",
            "54CD64AE-49FE-47D1-AF3A-8E7F02909E25": "iPhone 6s",
            "5773DBDD-EA60-4E53-8A96-57BD6C644DD5": "iPad Pro (9.7-inch)",
            "85A06C2A-ACDF-4A19-8C90-386DCF887670": "iPhone XS Max",
            "921C8E93-702E-47AB-A233-88982C2BBD95": "iPhone 6",
            "9F055949-5DF2-40D8-A955-A8517F213E24": "iPhone 8 Plus",
            "A1970E36-8906-48FF-8B3C-819A0A88D9D6": "iPhone 6 Plus",
            "A604E87D-B2BF-4190-B974-C29FC40A6F15": "iPhone 7 Plus",
            "AAF12280-0DC2-472F-87C5-2F141A6F0C55": "iPad Pro (12.9-inch) (2nd generation)",
            "AC8D34EB-F42D-4518-A09C-3C3AD7FCAC8C": "iPhone SE",
            "C4DE0942-3A85-4091-98CC-C4A90E2D07C3": "iPhone X",
            "C8E5AD6A-B7EB-480F-89E8-341FD45AAFFC": "iPad Air",
            "D675DECD-F18D-4950-A72C-22D2F3D39FFC": "iPad Pro (12.9-inch)",
            "DCBB1DBB-D3AF-4A65-AC35-A367B0E5BB8A": "iPad (6th generation)",
            "E98A6654-E843-43A5-8BD8-5D4C891EBA15": "iPhone 5s",
        },
        "iOS 12.1": {
            "01A9DC6A-58D3-4D7D-AC98-9F0689990DC6": "iPhone 8 Plus",
            "04325672-C35F-4E5E-BD08-EAC478B7165C": "iPhone XS",
            "28F0335D-1B4D-4493-A5C5-4E86E2916178": "iPhone 6",
            "28F16D36-8878-489F-A8CF-33E7037D252B": "iPad Pro (9.7-inch)",
            "512E11C5-5654-4F10-98D7-F75C50DF5DB7": "iPad Pro (12.9-inch) (3rd generation)",
            "53D7FAF6-83D7-415D-A3B4-20A9D8C37C44": "iPhone 5s",
            "5EF8EAA5-9D63-4F53-8896-57F9D59DECF9": "iPhone 6s",
            "61D96B3A-3747-41AC-92F7-2177E467A196": "iPad Pro (10.5-inch)",
            "64467336-1571-403C-9225-77133D27E525": "iPhone 7",
            "64B63780-6911-4459-8024-F97A2DA5E36D": "iPhone 7 Plus",
            "765EA3E6-5E8E-4792-BD8B-AEDC20FFAFDB": "iPad (6th generation)",
            "91F106B8-5928-45F4-B840-25F5496680BF": "iPhone X",
            "93F5DA98-21F3-4923-A403-B8433558CAFA": "iPhone 8",
            "95ACBE11-CB06-4EBB-9AC9-CCEE2AEB6901": "iPhone XR",
            "9E20D123-3F2B-4203-8B4C-78EFF946E303": "iPhone 6s Plus",
            "A5298485-C7DB-4D00-9B6D-1AC436BD3B1A": "iPad (5th generation)",
            "A8296054-BDAF-4EBF-A964-CFF0A528ED9C": "iPad Pro (12.9-inch)",
            "C07C5003-9728-4184-B030-063074C4972F": "iPhone SE",
            "CC3A7ECB-1BB2-4B77-9473-9ACCC06FC002": "iPhone XS Max",
            "D637BC6D-A53F-4E78-BDA7-FA0D59303350": "iPad Pro (11-inch)",
            "DC08D810-B9AD-4423-972E-3EE8949BC1F2": "iPad Air",
            "DEE6AF0E-596D-4713-8B57-8C77D45EED80": "iPad Air 2",
            "F022D86A-E404-46C0-98B2-9AB63AD7008B": "iPad Pro (12.9-inch) (2nd generation)",
            "F7EF0E11-864C-42A2-8D80-4DBE78AFD86B": "iPhone 6 Plus",
        },
    }

    # The prompt was not shown
    mock_input.assert_not_called()



import subprocess
from pathlib import Path

import pytest

from briefcase.exceptions import BriefcaseCommandError


def devices_result(name):
    """Load an adb devices result file from the sample directory, and return the
    content."""
    adb_samples = Path(__file__).parent / "devices"
    with (adb_samples / (name + ".out")).open(encoding="utf-8") as adb_output_file:
        return adb_output_file.read()


def test_no_devices(mock_tools, android_sdk):
    """If there are no devices, an empty list is returned."""
    mock_tools.subprocess.check_output.return_value = devices_result("no_devices")

    assert android_sdk.devices() == {}


def test_no_model(mock_tools, android_sdk):
    """If there is no model, return Unknown device (no model name)"""
    mock_tools.subprocess.check_output.return_value = devices_result("no_model")

    assert android_sdk.devices() == {
        "emulator-5554": {"name": "Unknown device (no model name)", "authorized": True}
    }


def test_one_emulator(mock_tools, android_sdk):
    """If there is a single emulator, it is returned."""
    mock_tools.subprocess.check_output.return_value = devices_result("one_emulator")

    assert android_sdk.devices() == {
        "emulator-5554": {
            "name": "Android SDK built for x86",
            "authorized": True,
        },
    }


def test_multiple_devices(mock_tools, android_sdk):
    """If there are multiple devices, they are all returned."""
    mock_tools.subprocess.check_output.return_value = devices_result("multiple_devices")

    assert android_sdk.devices() == {
        "041234567892009a": {
            "name": (
                "Device not available for development "
                "(unauthorized usb:336675328X transport_id:2)"
            ),
            "authorized": False,
        },
        "KABCDABCDA1513": {
            "name": "Kogan Agora 9",
            "authorized": True,
        },
        "emulator-5554": {
            "name": "Android SDK built for x86",
            "authorized": True,
        },
        "emulator-5556": {
            "name": "Unknown device (offline)",
            "authorized": False,
        },
    }


def test_adb_error(mock_tools, android_sdk):
    """If there is a problem invoking adb, an error is returned."""
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        returncode=69, cmd="adb devices -l"
    )

    with pytest.raises(BriefcaseCommandError):
        android_sdk.devices()


def test_daemon_start(mock_tools, android_sdk):
    """If ADB outputs the daemon startup message, ignore those messages."""
    mock_tools.subprocess.check_output.return_value = devices_result("daemon_start")

    assert android_sdk.devices() == {}


def test_physical_device_macOS(mock_tools, android_sdk):
    """An extra piece of device detail is returned on macOS for physical devices."""
    mock_tools.subprocess.check_output.return_value = devices_result(
        "physical_device_macOS"
    )

    assert android_sdk.devices() == {
        "200ABCDEFGHIJK": {
            "authorized": True,
            "name": "Pixel 7",
        }
    }


def test_device_permissions(mock_tools, android_sdk):
    """If AndroidSDK doesn't have access to the device, the error message can be
    parsed."""
    mock_tools.subprocess.check_output.return_value = devices_result("no_permissions")

    assert android_sdk.devices() == {
        "200ABCDEFGHIJK": {
            "authorized": False,
            "name": (
                "Device not available for development (no permissions "
                "(user russell is not in the plugdev group); "
                "see [http://developer.android.com/tools/device.html] "
                "usb:5-4.4.1 transport_id:1)"
            ),
        },
        "300ABCDEFGHIJK": {
            "authorized": False,
            "name": (
                "Device not available for development (no permissions "
                "(missing udev rules? user is in the plugdev group); "
                "see [http://developer.android.com/tools/device.html] "
                "usb:5-4.4.1 transport_id:1)"
            ),
        },
    }



import os
import subprocess
import sys
import time
from unittest import mock

import pytest

from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.macOS.app import macOSAppCreateCommand

from ....utils import create_file, create_installed_package, mock_tgz_download


@pytest.fixture
def create_command(tmp_path, first_app_templated):
    command = macOSAppCreateCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )

    # mock subprocess app context for this app
    command.tools[first_app_templated].app_context = mock.MagicMock(spec_set=Subprocess)

    return command


@pytest.mark.parametrize(
    "permissions, info, entitlements, context",
    [
        # No permissions
        (
            {},
            {},
            {},
            {
                "info": {},
                "entitlements": {
                    "com.apple.security.cs.allow-unsigned-executable-memory": True,
                    "com.apple.security.cs.disable-library-validation": True,
                },
            },
        ),
        # Only custom permissions
        (
            {},
            {
                "NSCustomPermission": "Custom message",
            },
            {
                "com.apple.vm.networking": True,
            },
            {
                "info": {
                    "NSCustomPermission": "Custom message",
                },
                "entitlements": {
                    "com.apple.security.cs.allow-unsigned-executable-memory": True,
                    "com.apple.security.cs.disable-library-validation": True,
                    "com.apple.vm.networking": True,
                },
            },
        ),
        # Camera permissions
        (
            {
                "camera": "I need to see you",
            },
            {},
            {},
            {
                "info": {
                    "NSCameraUsageDescription": "I need to see you",
                },
                "entitlements": {
                    "com.apple.security.cs.allow-unsigned-executable-memory": True,
                    "com.apple.security.cs.disable-library-validation": True,
                    "com.apple.security.device.camera": True,
                },
            },
        ),
        # Microphone permissions
        (
            {
                "microphone": "I need to hear you",
            },
            {},
            {},
            {
                "info": {
                    "NSMicrophoneUsageDescription": "I need to hear you",
                },
                "entitlements": {
                    "com.apple.security.cs.allow-unsigned-executable-memory": True,
                    "com.apple.security.cs.disable-library-validation": True,
                    "com.apple.security.device.microphone": True,
                },
            },
        ),
        # Coarse location permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
            },
            {},
            {},
            {
                "info": {
                    "NSLocationUsageDescription": "I need to know roughly where you are",
                },
                "entitlements": {
                    "com.apple.security.cs.allow-unsigned-executable-memory": True,
                    "com.apple.security.cs.disable-library-validation": True,
                    "com.apple.security.personal-information.location": True,
                },
            },
        ),
        # Fine location permissions
        (
            {
                "fine_location": "I need to know exactly where you are",
            },
            {},
            {},
            {
                "info": {
                    "NSLocationUsageDescription": "I need to know exactly where you are",
                },
                "entitlements": {
                    "com.apple.security.cs.allow-unsigned-executable-memory": True,
                    "com.apple.security.cs.disable-library-validation": True,
                    "com.apple.security.personal-information.location": True,
                },
            },
        ),
        # Background location permissions
        (
            {
                "background_location": "I always need to know where you are",
            },
            {},
            {},
            {
                "info": {
                    "NSLocationUsageDescription": "I always need to know where you are",
                },
                "entitlements": {
                    "com.apple.security.cs.allow-unsigned-executable-memory": True,
                    "com.apple.security.cs.disable-library-validation": True,
                    "com.apple.security.personal-information.location": True,
                },
            },
        ),
        # Coarse location background permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
                "background_location": "I always need to know where you are",
            },
            {},
            {},
            {
                "info": {
                    "NSLocationUsageDescription": "I always need to know where you are",
                },
                "entitlements": {
                    "com.apple.security.cs.allow-unsigned-executable-memory": True,
                    "com.apple.security.cs.disable-library-validation": True,
                    "com.apple.security.personal-information.location": True,
                },
            },
        ),
        # Fine location background permissions
        (
            {
                "fine_location": "I need to know exactly where you are",
                "background_location": "I always need to know where you are",
            },
            {},
            {},
            {
                "info": {
                    "NSLocationUsageDescription": "I always need to know where you are",
                },
                "entitlements": {
                    "com.apple.security.cs.allow-unsigned-executable-memory": True,
                    "com.apple.security.cs.disable-library-validation": True,
                    "com.apple.security.personal-information.location": True,
                },
            },
        ),
        # Coarse and fine location permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
                "fine_location": "I need to know exactly where you are",
            },
            {},
            {},
            {
                "info": {
                    "NSLocationUsageDescription": "I need to know exactly where you are",
                },
                "entitlements": {
                    "com.apple.security.cs.allow-unsigned-executable-memory": True,
                    "com.apple.security.cs.disable-library-validation": True,
                    "com.apple.security.personal-information.location": True,
                },
            },
        ),
        # Coarse and fine background location permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
                "fine_location": "I need to know exactly where you are",
                "background_location": "I always need to know where you are",
            },
            {},
            {},
            {
                "info": {
                    "NSLocationUsageDescription": "I always need to know where you are",
                },
                "entitlements": {
                    "com.apple.security.cs.allow-unsigned-executable-memory": True,
                    "com.apple.security.cs.disable-library-validation": True,
                    "com.apple.security.personal-information.location": True,
                },
            },
        ),
        # Photo library permissions
        (
            {
                "photo_library": "I need to see your library",
            },
            {},
            {},
            {
                "info": {
                    "NSPhotoLibraryUsageDescription": "I need to see your library",
                },
                "entitlements": {
                    "com.apple.security.cs.allow-unsigned-executable-memory": True,
                    "com.apple.security.cs.disable-library-validation": True,
                    "com.apple.security.personal-information.photo_library": True,
                },
            },
        ),
        # Override and augment by cross-platform definitions
        (
            {
                "fine_location": "I need to know where you are",
            },
            {
                "NSCustomMessage": "Custom message",
                "NSLocationUsageDescription": "Platform specific",
            },
            {
                "com.apple.security.personal-information.location": False,
                "com.apple.security.cs.disable-library-validation": False,
                "com.apple.vm.networking": True,
            },
            {
                "info": {
                    "NSLocationUsageDescription": "Platform specific",
                    "NSCustomMessage": "Custom message",
                },
                "entitlements": {
                    "com.apple.security.cs.allow-unsigned-executable-memory": True,
                    "com.apple.security.cs.disable-library-validation": False,
                    "com.apple.security.personal-information.location": False,
                    "com.apple.vm.networking": True,
                },
            },
        ),
    ],
)
def test_permissions_context(
    create_command, first_app, permissions, info, entitlements, context
):
    """Platform-specific permissions can be added to the context."""
    # Set the permission, info and entitlement values
    first_app.permission = permissions
    first_app.info = info
    first_app.entitlement = entitlements
    # Extract the cross-platform permissions
    x_permissions = create_command._x_permissions(first_app)
    # Check that the final platform permissions are rendered as expected.
    assert context == create_command.permissions_context(first_app, x_permissions)


def test_install_app_resources(create_command, first_app_templated, tmp_path):
    """The app bundle's modification time is updated when app resources are
    installed."""
    # Get the initial app modification time
    initial_timestamp = os.path.getmtime(
        create_command.binary_path(first_app_templated)
    )

    # Add a small sleep to make sure that a touch will definitely result in an updated
    # modification time
    time.sleep(0.1)

    # Install resources
    create_command.install_app_resources(first_app_templated)

    # Modification time has been updated, and is newer
    assert (
        os.path.getmtime(create_command.binary_path(first_app_templated))
        > initial_timestamp
    )


@pytest.mark.parametrize(
    "host_arch, other_arch",
    [
        ("arm64", "x86_64"),
        ("arm64", "x86_64"),
    ],
)
def test_install_app_packages(
    create_command,
    first_app_templated,
    tmp_path,
    host_arch,
    other_arch,
):
    """A 2-pass install of app packages is performed."""
    bundle_path = tmp_path / "base_path/build/first-app/macos/app"

    create_command.tools.host_arch = host_arch
    first_app_templated.requires = ["first", "second==1.2.3", "third>=3.2.1"]

    # Mock the result of finding the binary packages - 2 of the packages are binary;
    # the version on the loosely specified package doesn't match the lower bound.
    create_command.find_binary_packages = mock.Mock(
        return_value=[
            ("second", "1.2.3"),
            ("third", "3.4.5"),
        ]
    )

    # Mock the thin command so we can confirm it was invoked.
    create_command.thin_app_packages = mock.Mock()

    # Mock the merge command so we can confirm it was invoked.
    create_command.merge_app_packages = mock.Mock()

    create_command.install_app_requirements(first_app_templated, test_mode=False)

    # We looked for binary packages in the host app_packages
    create_command.find_binary_packages.assert_called_once_with(
        bundle_path / f"app_packages.{host_arch}",
        universal_suffix="_universal2",
    )

    # A request was made to install requirements
    assert create_command.tools[first_app_templated].app_context.run.mock_calls == [
        # First call is to install the initial packages on the host arch
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--no-python-version-warning",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / ('app_packages.' + host_arch)}",
                "first",
                "second==1.2.3",
                "third>=3.2.1",
            ],
            check=True,
            encoding="UTF-8",
        ),
        # Second call installs the binary packages for the other architecture.
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--no-python-version-warning",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / ('app_packages.' + other_arch)}",
                "--no-deps",
                "--only-binary",
                ":all:",
                "second==1.2.3",
                "third==3.4.5",
            ],
            check=True,
            encoding="UTF-8",
            env={
                "PYTHONPATH": str(
                    bundle_path / "support/platform-site" / f"macosx.{other_arch}"
                )
            },
        ),
    ]

    # The app packages folder has been created. The existence of the target and host
    # versions is validated as a result of the underlying install/merge methods.
    assert (bundle_path / f"app_packages.{other_arch}").is_dir()

    # An attempt was made thin the "other" arch packages.
    create_command.thin_app_packages.assert_called_once_with(
        bundle_path / f"app_packages.{other_arch}",
        arch=other_arch,
    )

    # An attempt was made to merge packages.
    create_command.merge_app_packages.assert_called_once_with(
        target_app_packages=bundle_path
        / "First App.app"
        / "Contents"
        / "Resources"
        / "app_packages",
        sources=[
            bundle_path / f"app_packages.{host_arch}",
            bundle_path / f"app_packages.{other_arch}",
        ],
    )


@pytest.mark.parametrize(
    "host_arch, other_arch",
    [
        ("arm64", "x86_64"),
        ("arm64", "x86_64"),
    ],
)
def test_install_app_packages_no_binary(
    create_command,
    first_app_templated,
    tmp_path,
    host_arch,
    other_arch,
):
    """If there's no binaries in the first pass, the second pass isn't performed."""
    bundle_path = tmp_path / "base_path/build/first-app/macos/app"

    # Create pre-existing other-arch content
    create_installed_package(bundle_path / f"app_packages.{other_arch}", "legacy")

    create_command.tools.host_arch = host_arch
    first_app_templated.requires = ["first", "second==1.2.3", "third>=3.2.1"]

    # Mock the result of finding no binary packages.
    create_command.find_binary_packages = mock.Mock(return_value=[])

    # Mock the thin command so we can confirm it was invoked.
    create_command.thin_app_packages = mock.Mock()

    # Mock the merge command so we can confirm it was invoked.
    create_command.merge_app_packages = mock.Mock()

    create_command.install_app_requirements(first_app_templated, test_mode=False)

    # We looked for binary packages in the host app_packages
    create_command.find_binary_packages.assert_called_once_with(
        bundle_path / f"app_packages.{host_arch}",
        universal_suffix="_universal2",
    )

    # A request was made to install requirements
    assert create_command.tools[first_app_templated].app_context.run.mock_calls == [
        # Only call is to install the initial packages on the host arch
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--no-python-version-warning",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / ('app_packages.' + host_arch)}",
                "first",
                "second==1.2.3",
                "third>=3.2.1",
            ],
            check=True,
            encoding="UTF-8",
        ),
    ]

    # The app packages folder for the other architecture has been created, even though
    # it isn't needed. The existence of the target and host versions is validated as a
    # result of the underlying install/merge methods.
    assert (bundle_path / f"app_packages.{other_arch}").is_dir()

    # We still need to thin and merge the app packages; this is effectively just a copy.
    create_command.thin_app_packages.assert_called_once_with(
        bundle_path / f"app_packages.{other_arch}",
        arch=other_arch,
    )
    create_command.merge_app_packages.assert_called_once_with(
        target_app_packages=bundle_path
        / "First App.app"
        / "Contents"
        / "Resources"
        / "app_packages",
        sources=[
            bundle_path / f"app_packages.{host_arch}",
            bundle_path / f"app_packages.{other_arch}",
        ],
    )


def test_install_app_packages_failure(create_command, first_app_templated, tmp_path):
    """If the install of other-arch binaries fails, an exception is raised."""
    bundle_path = tmp_path / "base_path/build/first-app/macos/app"

    # Create pre-existing other-arch content
    create_installed_package(bundle_path / "app_packages.x86_64", "legacy")

    create_command.tools.host_arch = "arm64"
    first_app_templated.requires = ["first", "second==1.2.3", "third>=3.2.1"]

    # Mock the result of finding the binary packages - 2 of the packages are binary;
    # the version on the loosely specified package doesn't match the lower bound.
    create_command.find_binary_packages = mock.Mock(
        return_value=[
            ("second", "1.2.3"),
            ("third", "3.4.5"),
        ]
    )

    # Mock the thin command so we can confirm it was invoked.
    create_command.thin_app_packages = mock.Mock()

    # Mock the merge command so we can confirm it was invoked.
    create_command.merge_app_packages = mock.Mock()

    # Mock a failure on the second install
    create_command.tools[first_app_templated].app_context.run.side_effect = [
        None,
        subprocess.CalledProcessError(returncode=1, cmd="pip"),
    ]

    # Install the requirements; this will raise an error
    with pytest.raises(
        BriefcaseCommandError,
        match=(
            r"Unable to install requirements\. This may be because one of your\n"
            r"requirements is invalid, or because pip was unable to connect\n"
            r"to the PyPI server.\n"
        ),
    ):
        create_command.install_app_requirements(first_app_templated, test_mode=False)

    # We looked for binary packages in the host app_packages
    create_command.find_binary_packages.assert_called_once_with(
        bundle_path / "app_packages.arm64",
        universal_suffix="_universal2",
    )

    # A request was made to install requirements
    assert create_command.tools[first_app_templated].app_context.run.mock_calls == [
        # First call is to install the initial packages on the host arch
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--no-python-version-warning",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / 'app_packages.arm64'}",
                "first",
                "second==1.2.3",
                "third>=3.2.1",
            ],
            check=True,
            encoding="UTF-8",
        ),
        # Second call installs the binary packages for the other architecture;
        # this is the call that failed.
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--no-python-version-warning",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / 'app_packages.x86_64'}",
                "--no-deps",
                "--only-binary",
                ":all:",
                "second==1.2.3",
                "third==3.4.5",
            ],
            check=True,
            encoding="UTF-8",
            env={
                "PYTHONPATH": str(bundle_path / "support/platform-site/macosx.x86_64")
            },
        ),
    ]

    # The app packages folder for the other architecture has been created, even though
    # it isn't needed. The existence of the target and host versions is validated as a
    # result of the underlying install/merge methods.
    assert (bundle_path / "app_packages.x86_64").is_dir()

    # We didn't attempt to thin or  merge, because we didn't complete installing.
    create_command.thin_app_packages.assert_not_called()
    create_command.merge_app_packages.assert_not_called()


@pytest.mark.parametrize(
    "host_arch, other_arch",
    [
        ("arm64", "x86_64"),
        ("arm64", "x86_64"),
    ],
)
def test_install_app_packages_non_universal(
    create_command,
    first_app_templated,
    tmp_path,
    host_arch,
    other_arch,
):
    """If the app is non-universal, only a single install pass occurs, followed by
    thinning."""
    bundle_path = tmp_path / "base_path/build/first-app/macos/app"

    create_command.tools.host_arch = host_arch
    first_app_templated.requires = ["first", "second==1.2.3", "third>=3.2.1"]
    first_app_templated.universal_build = False

    # Mock the find_binary_packages command so we can confirm it wasn't invoked.
    create_command.find_binary_packages = mock.Mock()

    # Mock the thin command so we can confirm it was invoked.
    create_command.thin_app_packages = mock.Mock()

    # Mock the merge command so we can confirm it wasn't invoked.
    create_command.merge_app_packages = mock.Mock()

    create_command.install_app_requirements(first_app_templated, test_mode=False)

    # We didn't search for binary packages
    create_command.find_binary_packages.assert_not_called()

    # One request was made to install requirements
    assert create_command.tools[first_app_templated].app_context.run.mock_calls == [
        mock.call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--no-python-version-warning",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / 'First App.app' / 'Contents' / 'Resources' / 'app_packages'}",
                "first",
                "second==1.2.3",
                "third>=3.2.1",
            ],
            check=True,
            encoding="UTF-8",
        ),
    ]

    # An attempt was made to thin the app packages
    create_command.thin_app_packages.assert_called_once_with(
        bundle_path / "First App.app/Contents/Resources/app_packages",
        arch=host_arch,
    )

    # No attempt was made to merge packages.
    create_command.merge_app_packages.assert_not_called()


@pytest.mark.parametrize("pre_existing", [True, False])
def test_install_support_package(
    create_command,
    first_app_templated,
    tmp_path,
    pre_existing,
):
    """The standard library is copied out of the support package into the app bundle."""
    # Hard code the support revision
    first_app_templated.support_revision = "37"

    bundle_path = tmp_path / "base_path/build/first-app/macos/app"
    runtime_support_path = bundle_path / "First App.app/Contents/Resources/support"

    if pre_existing:
        create_file(
            runtime_support_path / "python-stdlib/old-stdlib",
            "Legacy stdlib file",
        )

    # Mock download.file to return a support package
    create_command.tools.file.download = mock.MagicMock(
        side_effect=mock_tgz_download(
            f"Python-3.{sys.version_info.minor}-macOS-support.b37.tar.gz",
            [
                ("python-stdlib/stdlib.txt", "this is the standard library"),
                (
                    "platform-site/macosx.arm64/sitecustomize.py",
                    "this is the arm64 platform site",
                ),
                (
                    "platform-site/macosx.x86_64/sitecustomize.py",
                    "this is the x86_64 platform site",
                ),
                ("Python.xcframework/info.plist", "this is the xcframework"),
            ],
        )
    )

    # Install the support package
    create_command.install_app_support_package(first_app_templated)

    # Confirm that the support files have been unpacked into the bundle location
    assert (bundle_path / "support/python-stdlib/stdlib.txt").exists()
    assert (
        bundle_path / "support/platform-site/macosx.arm64/sitecustomize.py"
    ).exists()
    assert (
        bundle_path / "support/platform-site/macosx.x86_64/sitecustomize.py"
    ).exists()
    assert (bundle_path / "support/Python.xcframework/info.plist").exists()

    # The standard library has been copied to the app...
    assert (runtime_support_path / "python-stdlib/stdlib.txt").exists()
    # ... but the other support files have not.
    assert not (
        runtime_support_path / "platform-site/macosx.arm64/sitecustomize.py"
    ).exists()
    assert not (
        runtime_support_path / "platform-site/macosx.x86_64/sitecustomize.py"
    ).exists()
    assert not (runtime_support_path / "Python.xcframework/info.plist").exists()

    # The legacy content has been purged
    assert not (runtime_support_path / "python-stdlib/old-stdlib").exists()



from __future__ import annotations

import concurrent
import email
import hashlib
import subprocess
from pathlib import Path

from briefcase.exceptions import BriefcaseCommandError


def sha256_file_digest(path: Path) -> str:
    """Compute a sha256 checksum digest of a file.

    :param path: The file to digest.
    :returns: A sha256 hex digest for the file.
    """
    with path.open("rb") as f:
        file_hash = hashlib.sha256()
        while chunk := f.read(8192):
            file_hash.update(chunk)

    return file_hash.hexdigest()


class AppPackagesMergeMixin:
    # A mixin containing the utilities to merge independent platform-specific app_packages folders
    # into a single "fat" app_packages folder. This is currently only used by macOS, but it *could*
    # be required on iOS if they ever re-introduce multiple on-device architectures.

    def find_binary_packages(
        self,
        install_path: Path,
        universal_suffix: str = None,
    ) -> list[tuple[str, str]]:
        """Find the packages that have been installed that have binary components.

        :param install_path: The path into which packages have been installed.
        :param universal_suffix: The tag suffix that indicates a universal wheel.
        :returns: A list of (package name, version) tuples, describing the packages that
            are in the install path that are non-universal and non-pure.
        """
        binary_packages = []
        for distinfo in install_path.glob("**/*.dist-info"):
            # Read the WHEEL file in the dist-info folder.
            # Use this to determine if the wheel is "pure", and the tag
            # for the wheel.
            with (distinfo / "WHEEL").open("r", encoding="utf-8") as f:
                wheel_data = email.message_from_string(f.read())
                is_purelib = wheel_data.get("Root-Is-Purelib", "false") == "true"
                tag = wheel_data["Tag"]

            # If the wheel is pure, it's not a binary package
            if is_purelib:
                continue

            # If the tag ends with the universal tag, the binary package can be used on all
            # targets, and doesn't need additional processing.
            if universal_suffix and tag.endswith(universal_suffix):
                continue

            # The wheel is a single platform binary wheel.
            # Read the package metadata to determine what needs to be installed.
            with (distinfo / "METADATA").open("r", encoding="utf-8") as f:
                metadata = email.message_from_string(f.read())
                binary_packages.append((metadata["Name"], metadata["Version"]))

        return binary_packages

    def ensure_thin_binary(self, path: Path, arch: str):
        """Ensure that a binary is thin, targeting a given architecture.

        If the library is already thin, it is left as-is.

        :param path: The library file to process.
        :param arch: The architecture that should be preserved.
        """
        try:
            output = self.tools.subprocess.check_output(
                ["lipo", "-info", path],
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Unable to inspect architectures in {path}"
            ) from e
        else:
            if output.startswith("Non-fat file: "):
                self.logger.verbose(f"{path} is already thin.")
            elif output.startswith("Architectures in the fat file: "):
                architectures = set(output.strip().split(":")[-1].strip().split(" "))
                if arch in architectures:
                    self.logger.verbose(f"Thinning {path}")
                    try:
                        thin_lib_path = path.parent / f"{path.name}.{arch}"
                        self.tools.subprocess.run(
                            [
                                "lipo",
                                "-thin",
                                arch,
                                "-output",
                                thin_lib_path,
                                path,
                            ],
                            check=True,
                        )
                    except subprocess.CalledProcessError as e:
                        raise BriefcaseCommandError(
                            f"Unable to create thin binary from {path}"
                        ) from e
                    else:
                        # Having extracted the single architecture into a temporary
                        # file, replace the original with the thin version.
                        self.tools.shutil.move(thin_lib_path, path)
                else:
                    raise BriefcaseCommandError(
                        f"{path} does not contain a {arch} slice."
                    )
            else:
                raise BriefcaseCommandError(
                    f"Unable to determine architectures in {path}"
                )

    def lipo_dylib(self, relative_path: Path, target_path: Path, sources: list[Path]):
        """Create a fat library by invoking lipo on multiple source libraries.

        :param relative_path: The path fragment for the dylib, relative to the root
        :param target_path: The root location where the fat library will be written
        :param sources: A list of root locations providing single platform libraries.
        """
        self.logger.verbose(f"Creating fat library {relative_path}")

        try:
            # Ensure the directory where the library will be written exists.
            (target_path / relative_path).parent.mkdir(exist_ok=True, parents=True)

            # Add all the constructed source paths. If the original binary is universal,
            # or the binary is only needed on *some* platforms (e.g., libjpeg isn't
            # included in the x86_64 Pillow wheel), the source won't exist, so only
            # merge sources that actually exist. lipo allows creating a "fat"
            # single-platform binary; it's effectively a copy.
            self.tools.subprocess.run(
                [
                    "lipo",
                    "-create",
                    "-output",
                    target_path / relative_path,
                ]
                + [
                    source_path / relative_path
                    for source_path in sources
                    if (source_path / relative_path).is_file()
                ],
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                f"Unable to create fat library for {relative_path}"
            ) from e

    def thin_app_packages(
        self,
        app_packages: Path,
        arch: str,
    ):
        """Ensure that all the dylibs in a given app_packages folder are thin."""
        dylibs = []
        for source_path in app_packages.glob("**/*"):
            if source_path.suffix in {".so", ".dylib"}:
                dylibs.append(source_path)

        # Call lipo on each dylib that was found to ensure it is thin.
        if dylibs:
            # Do this in a threadpool to make it run faster.
            progress_bar = self.input.progress_bar()
            self.logger.info(f"Thinning libraries in {app_packages.name}...")
            task_id = progress_bar.add_task("Create fat libraries", total=len(dylibs))
            with progress_bar:
                with concurrent.futures.ThreadPoolExecutor() as executor:
                    futures = []
                    for path in dylibs:
                        future = executor.submit(
                            self.ensure_thin_binary,
                            path=path,
                            arch=arch,
                        )
                        futures.append(future)
                    for future in concurrent.futures.as_completed(futures):
                        progress_bar.update(task_id, advance=1)
                        if future.exception():
                            raise future.exception()
        else:
            self.logger.info("No libraries require thinning.")

    def merge_app_packages(
        self,
        target_app_packages: Path,
        sources: list[Path],
    ):
        """Merge the contents of multiple source app packages into a final location.

        :param target_app_packages: The location of the final merged app packages
        :param sources: A list of app packages folders to be merged.
        """
        # Clear any existing app packages folder.
        if target_app_packages.is_dir():
            self.tools.shutil.rmtree(target_app_packages)
        self.tools.os.mkdir(target_app_packages)

        # Copy all the non-library files from the source to the target. If a file exists
        # in multiple sources, do a checksum, and warn if there is a discrepancy. If
        # there is a discrepancy, the version from the first source is what will be
        # used. Don't warn if the file is in a __pycache__ folder; these will always be
        # different, but they'll be purged later anyway. Don't warn for anything in a
        # .dist-info folder either; these are going to be different because of platform
        # difference, but core package metadata should be consistent.
        dylibs = set()
        digests = {}
        for source_app_packages in sources:
            with self.input.wait_bar(f"Merging {source_app_packages.name}..."):
                for source_path in source_app_packages.glob("**/*"):
                    relative_path = source_path.relative_to(source_app_packages)
                    target_path = target_app_packages / relative_path
                    if source_path.is_dir():
                        target_path.mkdir(exist_ok=True)
                    else:
                        if source_path.suffix in {".so", ".dylib"}:
                            # Dynamic libraries need to be merged; do this in a second pass.
                            dylibs.add(relative_path)
                        elif target_path.exists():
                            # The file already exists. Check for differences; if there are any
                            # differences outside `dist-info` or `__pycache__` folders, warn
                            # the user.
                            digest = sha256_file_digest(source_path)
                            if digests[relative_path] != digest and not (
                                relative_path.parent.name == "__pycache__"
                                or Path(relative_path.parts[0]).suffix == ".dist-info"
                            ):
                                self.logger.warning(
                                    f"{relative_path} has different content "
                                    f"between sources; ignoring {source_app_packages.suffix[1:]} version."
                                )
                        else:
                            # The file doesn't exist yet; copy it as is (including
                            # permissions), and store the digest for later comparison
                            self.tools.shutil.copy(source_path, target_path)
                            digests[relative_path] = sha256_file_digest(source_path)

        # Call lipo on each dylib that was found to create the fat version.
        if dylibs:
            # Do this in a threadpool to make it run faster.
            progress_bar = self.input.progress_bar()
            self.logger.info("Merging libraries...")
            task_id = progress_bar.add_task("Create fat libraries", total=len(dylibs))
            with progress_bar:
                with concurrent.futures.ThreadPoolExecutor() as executor:
                    futures = []
                    for relative_path in dylibs:
                        future = executor.submit(
                            self.lipo_dylib,
                            target_path=target_app_packages,
                            relative_path=relative_path,
                            sources=sources,
                        )
                        futures.append(future)
                    for future in concurrent.futures.as_completed(futures):
                        progress_bar.update(task_id, advance=1)
                        if future.exception():
                            raise future.exception()
        else:
            self.logger.info("No libraries require merging.")



def test_run_app_context(mock_sub):
    """Run app context manager returns the input."""
    in_kwargs = object()
    with mock_sub.run_app_context(subprocess_kwargs=in_kwargs) as kwargs:
        assert in_kwargs is kwargs



from unittest.mock import MagicMock


def test_multiple_pep621_options(convert_command, monkeypatch):
    mock_select_option = MagicMock()
    mock_select_option.return_value = "https://some_other_url.com/"
    monkeypatch.setattr(convert_command, "select_option", mock_select_option)

    (convert_command.base_path / "pyproject.toml").write_text(
        "[project.urls]\n"
        'Homepage="https://some_url.com/"\n'
        'Documentation="https://some_other_url.com/"',
        encoding="utf-8",
    )
    convert_command.input_url("some-name", None)
    mock_select_option.assert_called_once_with(
        intro=(
            "What website URL do you want to use for this application? The "
            "following URLs are defined in your existing 'pyproject.toml'; "
            "select 'Other' to provide a different URL."
        ),
        variable="application URL",
        default=None,
        options=[
            "https://some_url.com/",
            "https://some_other_url.com/",
            "Other",
        ],
        override_value=None,
    )


def test_multiple_pep621_options_select_other(convert_command, monkeypatch):
    mock_select_option = MagicMock()
    mock_select_option.return_value = "Other"
    monkeypatch.setattr(convert_command, "select_option", mock_select_option)

    mock_input_text = MagicMock()
    monkeypatch.setattr(convert_command, "input_text", mock_input_text)

    (convert_command.base_path / "pyproject.toml").write_text(
        "[project.urls]\n"
        'Homepage="https://some_url.com/"\n'
        'Documentation="https://some_other_url.com/"',
        encoding="utf-8",
    )
    convert_command.input_url("some-name", None)
    mock_select_option.assert_called_once_with(
        intro=(
            "What website URL do you want to use for this application? The "
            "following URLs are defined in your existing 'pyproject.toml'; "
            "select 'Other' to provide a different URL."
        ),
        variable="application URL",
        default=None,
        options=[
            "https://some_url.com/",
            "https://some_other_url.com/",
            "Other",
        ],
        override_value=None,
    )
    mock_input_text.assert_called_once_with(
        intro="What website URL do you want to use for the application?",
        variable="application URL",
        default="https://example.com/some-name",
        validator=convert_command.validate_url,
    )


def test_no_pep621_options(convert_command, monkeypatch):
    mock_input_text = MagicMock()
    monkeypatch.setattr(convert_command, "input_text", mock_input_text)

    convert_command.input_url("some-name", None)
    default = "https://example.com/some-name"
    mock_input_text.assert_called_once_with(
        intro=(
            "What website URL do you want to use for this application? Based "
            f"on your existing 'pyproject.toml', this might be {default}"
        ),
        variable="application URL",
        default=default,
        validator=convert_command.validate_url,
        override_value=None,
    )


def test_override(convert_command):
    assert (
        convert_command.input_url("some-name", "https://example.com/override")
        == "https://example.com/override"
    )


def test_prompted_url(convert_command):
    """You can type in the URL."""
    convert_command.input.values = ["https://example.com/some-name"]
    assert (
        convert_command.input_url("some-name", None) == "https://example.com/some-name"
    )



from __future__ import annotations

import contextlib
import json
import operator
import os
import queue
import shlex
import subprocess
import sys
import threading
import time
from collections.abc import Callable
from functools import wraps
from pathlib import Path
from subprocess import CompletedProcess
from typing import Iterator, Mapping, Sequence, TypeVar, Union

import psutil

from briefcase.config import AppConfig
from briefcase.console import Log
from briefcase.exceptions import CommandOutputParseError, ParseError
from briefcase.integrations.base import Tool, ToolCache

SubprocessArgT = Union[str, Path]
SubprocessArgsT = Sequence[SubprocessArgT]
JsonT = Union[Mapping[str, "JsonT"], Sequence["JsonT"], str, int, float, bool, None]
ParserOutputT = TypeVar("ParserOutputT")


class StopStreaming(Exception):
    """Raised by streaming filters to terminate the stream."""


def ensure_str(text: str | bytes) -> str:
    """Returns input text as a string."""
    return text.decode() if isinstance(text, bytes) else str(text)


def json_parser(json_output: str) -> JsonT:
    """Wrapper to parse command output as JSON via parse_output.

    :param json_output: command output to parse as JSON
    """
    try:
        return json.loads(json_output)
    except json.JSONDecodeError as e:
        raise ParseError(f"Failed to parse output as JSON: {e}") from e


def is_process_dead(pid: int) -> bool:
    """Returns True if a PID is not assigned to a process.

    Checking if a PID exists is only a semi-safe proxy to determine if a process is dead
    since PIDs can be re-used. Therefore, this function should only be used via constant
    monitoring of a PID to identify when the process goes from existing to not existing.

    :param pid: integer value to be checked if assigned as a PID.
    :returns: True if PID does not exist; False otherwise.
    """
    return not psutil.pid_exists(pid)


def get_process_id_by_command(
    command_list: list[str] | None = None,
    command: str = "",
    logger: Log | None = None,
) -> int | None:
    """Find a Process ID (PID) a by its command. If multiple processes are found, then
    the most recently created process ID is returned.

    :param command_list: list of a command's fully qualified path and its arguments.
    :param command: a partial or complete fully-qualified filepath to a command.
        This is primarily intended for use on macOS where the `open` command
        takes a filepath to a directory for an application; therefore, the actual
        running process will be running a command within that directory.
    :param logger: optional Log to show messages about process matching to users
    :returns: PID if found else None
    """
    matching_procs = []
    # retrieve command line, creation time, and ID for all running processes.
    # note: psutil returns None for a process attribute if it is unavailable;
    #   this is most likely to happen for restricted or zombie processes.
    for proc in psutil.process_iter(["cmdline", "create_time", "pid"]):
        proc_cmdline: list[str] = proc.info["cmdline"]
        if command_list and proc_cmdline == command_list:
            matching_procs.append(proc.info)
        if command and proc_cmdline and proc_cmdline[0].startswith(command):
            matching_procs.append(proc.info)

    if len(matching_procs) == 1:
        return matching_procs[0]["pid"]
    elif len(matching_procs) > 1:
        # return the ID of the most recently created matching process
        pid = sorted(matching_procs, key=operator.itemgetter("create_time"))[-1]["pid"]
        if logger:
            logger.info(
                f"Multiple running instances of app found. Using most recently created app process {pid}."
            )
        return pid

    return None


def ensure_console_is_safe(sub_method):
    """Decorator for Subprocess methods to conditionally remove dynamic console elements
    such as the Wait Bar prior to running the subprocess command.

    :param sub_method: wrapped Subprocess method
    """

    @wraps(sub_method)
    def inner(sub: Subprocess, args: SubprocessArgsT, *wrapped_args, **wrapped_kwargs):
        """Evaluate whether conditions are met to remove any dynamic elements in the
        console before returning control to Subprocess.

        :param sub: Bound Subprocess object
        :param args: command line to run in subprocess
        :returns: the return value for the Subprocess method
        """
        # Just run the command if no dynamic elements are active
        if not sub.tools.input.is_console_controlled:
            return sub_method(sub, args, *wrapped_args, **wrapped_kwargs)

        remove_dynamic_elements = False

        # Batch (.bat) scripts on Windows.
        # If cmd.exe is interrupted with CTRL+C while running a bat script,
        # it may prompt the user to abort the script and dynamic elements
        # such as the Wait Bar can hide this message from the user.
        if sub.tools.host_os == "Windows":
            executable = str(args[0]).strip() if args else ""
            remove_dynamic_elements |= executable.lower().endswith(".bat")

        # Release control for commands that cannot be streamed.
        remove_dynamic_elements |= wrapped_kwargs.get("stream_output") is False

        # Run subprocess command with or without console control
        if remove_dynamic_elements:
            with sub.tools.input.release_console_control():
                return sub_method(sub, args, *wrapped_args, **wrapped_kwargs)
        else:
            return sub_method(sub, args, *wrapped_args, **wrapped_kwargs)

    return inner


class PopenOutputStreamer(threading.Thread):
    def __init__(
        self,
        label: str,
        popen_process: subprocess.Popen,
        logger: Log,
        capture_output: bool = False,
        filter_func: Callable[[str], Iterator[str]] | None = None,
    ):
        """Thread for streaming stdout for a Popen process.

        :param label: Descriptive name for process being streamed
        :param popen_process: Popen process with stdout to stream
        :param logger: logger for printing to console
        :param capture_output: Retain process output in ``output_queue`` via a
            ``queue.Queue`` instead of printing to console
        :param filter_func: a callable that will be invoked on every line of output
            that is streamed; see ``Subprocess.stream_output`` for details
        """
        super().__init__(name=f"{label} output streamer", daemon=True)

        self.popen_process = popen_process
        self.logger = logger
        self.capture_output = capture_output
        self.filter_func = filter_func

        # arbitrarily large maxsize to prevent unbounded memory use if things go south
        self.output_queue = queue.Queue(maxsize=10_000_000)
        self.stop_flag = threading.Event()

    def run(self):
        """Stream output for a Popen process."""
        try:
            while output_line := self._readline():
                # The stop_flag is intentionally checked both at the top and bottom of
                # this loop; if the flag was set during the call to readline(), then
                # processing the output is skipped altogether. And if the flag is set
                # as a consequence of filter_func(), the streamer still exits before
                # calling readline() again and potentially blocking indefinitely.
                if not self.stop_flag.is_set():
                    filtered_output, stop_streaming = self._filter(output_line)

                    for filtered_line in filtered_output:
                        if self.capture_output:
                            self.output_queue.put_nowait(filtered_line)
                        else:
                            self.logger.info(filtered_line)

                    if stop_streaming:
                        self.stop_flag.set()

                if self.stop_flag.is_set():
                    break
        except Exception as e:
            self.logger.error(f"Error while streaming output: {type(e).__name__}: {e}")
            self.logger.capture_stacktrace("Output thread")

    def request_stop(self):
        """Set the stop flag to cause the streamer to exit.

        If the streamer is currently blocking on `readline()` because the process'
        stdout buffer is empty, then the streamer will not exit until `readline()`
        returns or until Briefcase exits.
        """
        self.stop_flag.set()

    @property
    def captured_output(self) -> str:
        """The captured output from the process."""
        output = []
        while not self.output_queue.empty():
            with contextlib.suppress(queue.Empty):
                output.append(self.output_queue.get_nowait())
                self.output_queue.task_done()
        return "".join(output)

    def _readline(self) -> str:
        """Read a line of output from the process while blocking.

        Calling readline() for stdout always returns at least a newline, i.e. "\n",
        UNLESS the process is exiting or already exited; in that case, an empty string
        is returned.

        :returns: one line of output or "" if nothing more can be read from stdout
        """
        try:
            return ensure_str(self.popen_process.stdout.readline())
        except ValueError as e:
            # Catch ValueError if stdout is unexpectedly closed; this can
            # happen, for instance, if the user starts spamming CTRL+C.
            if "I/O operation on closed file" in str(e):
                self.logger.warning(
                    "WARNING: stdout was unexpectedly closed while streaming output"
                )
                return ""
            else:
                raise

    def _filter(self, line: str) -> tuple[list[str], bool]:
        """Run filter function over output from process."""
        filtered_output = []
        stop_streaming = False

        if self.filter_func is not None:
            try:
                for filtered_line in self.filter_func(line.strip("\n")):
                    filtered_output.append(filtered_line)
            except StopStreaming:
                stop_streaming = True
        else:
            filtered_output.append(line)

        return filtered_output, stop_streaming


class NativeAppContext(Tool):
    """A wrapper around subprocess for use as an app-bound tool."""

    name = "app_context_subprocess"

    @classmethod
    def verify_install(cls, tools: ToolCache, app: AppConfig, **kwargs) -> Subprocess:
        """Make subprocess available as app-bound tool."""
        # short circuit since already verified and available
        if hasattr(tools[app], "app_context"):
            return tools[app].app_context

        tools[app].app_context = tools.subprocess
        return tools[app].app_context


class Subprocess(Tool):
    """A wrapper around subprocess that can be used as a logging point for commands that
    are executed."""

    name = "subprocess"
    full_name = "Subprocess"

    def __init__(self, tools: ToolCache):
        super().__init__(tools=tools)
        self._subprocess = subprocess

    def prepare(self):
        """Perform any environment preparation required to execute processes."""
        # This is a no-op; the native subprocess environment is ready-to-use.
        pass

    @contextlib.contextmanager
    def run_app_context(self, subprocess_kwargs: dict[str, ...]) -> dict[str, ...]:
        """A manager to wrap subprocess calls to run a Briefcase project app.

        :param subprocess_kwargs: initialized keyword arguments for subprocess calls
        """
        # This is a no-op; the native subprocess environment is ready-to-use.
        yield subprocess_kwargs

    def full_env(self, overrides: dict[str, str | None] | None) -> dict[str, str]:
        """Generate the full environment in which the command will run.

        If an env var in `overrides` is set to `None`, then that env var
        will be altogether absent in the returned environment.

        :param overrides: The environment passed to the subprocess call;
            can be `None` if there are no explicit environment changes.
        """
        env = self.tools.os.environ.copy()
        if overrides:
            env.update(overrides)
            env = {k: v for k, v in env.items() if v is not None}
        return env

    def final_kwargs(self, **kwargs) -> dict[str, str]:
        """Convert subprocess keyword arguments into their final form.

        This involves:
         * Converting any environment overrides into a full environment
         * Converting the `cwd` into a string
         * Default `text` to True so all outputs are strings
         * Convert start_new_session=True to creationflags on Windows
        """
        # If `env` has been provided, inject a full copy of the local
        # environment, with the values in `env` overriding the local
        # environment.
        try:
            overrides = kwargs.pop("env")
            kwargs["env"] = self.full_env(overrides)
        except KeyError:
            # No explicit environment provided.
            pass

        # If `cwd` has been provided, ensure it is in string form.
        try:
            cwd = kwargs.pop("cwd")
            kwargs["cwd"] = str(cwd)
        except KeyError:
            pass

        # Default to running subprocess in "text" mode so all output is returned as
        # strings instead of bytes. The legacy setting of `universal_newlines` is
        # converted to `text` for calling subprocess.
        kwargs["text"] = kwargs.pop("universal_newlines", kwargs.get("text", True))

        # Enable text mode if an `encoding` or `errors` is specified; this aligns with
        # the default behavior of subprocess.
        kwargs["text"] |= any(kwargs.get(kw) for kw in ["encoding", "errors"])

        # Configure subprocess for text mode
        if kwargs["text"]:
            # Ensure an appropriate encoding is specified.
            # If an encoding isn't provided, default to the encoding that Python is
            # using for stdout. This is for the benefit of Windows which uses cp437 for
            # console output when the system encoding is cp1252.
            # `sys.__stdout__` is used because Rich captures and redirects `sys.stdout`.
            # The fallback to "UTF-8" is needed to catch the case where stdout is
            # redirected to a non-tty (e.g. during pytest conditions).
            kwargs.setdefault(
                "encoding", os.device_encoding(sys.__stdout__.fileno()) or "UTF-8"
            )

            # Use relaxed output decoding by default.
            # subprocess defaults to "strict" handling for errors arising from decoding
            # output to Unicode. To avoid Unicode exceptions from misbehaving commands,
            # set `errors` so output that cannot be decoded for the specified encoding
            # are replaced with hex of the raw bytes.
            kwargs.setdefault("errors", "backslashreplace")

        # For Windows, convert start_new_session=True to creation flags
        if self.tools.host_os == "Windows":
            try:
                if kwargs.pop("start_new_session") is True:
                    if "creationflags" in kwargs:
                        raise AssertionError(
                            "Subprocess called with creationflags set and "
                            "start_new_session=True.\nThis will result in "
                            "CREATE_NEW_PROCESS_GROUP and CREATE_NO_WINDOW "
                            "being merged in to the creationflags.\n\nEnsure "
                            "this is desired configuration or don't set "
                            "start_new_session=True."
                        )
                    # CREATE_NEW_PROCESS_GROUP: Promotes the new process to the root
                    #    process of a new process group. This also disables CTRL+C
                    #    signal handlers for all processes of the new process group.
                    # CREATE_NO_WINDOW: Creates a new console for the process but
                    #    does not open a visible window for that console. This flag
                    #    is used instead of DETACHED_PROCESS since the new process
                    #    can spawn a new console itself (in the absence of one being
                    #    available) but that console creation will also spawn a
                    #    visible console window.
                    new_session_flags = (
                        self._subprocess.CREATE_NEW_PROCESS_GROUP
                        | self._subprocess.CREATE_NO_WINDOW
                    )
                    # merge these flags in to any existing flags already provided
                    kwargs["creationflags"] = (
                        kwargs.get("creationflags", 0) | new_session_flags
                    )
            except KeyError:
                pass

        return kwargs

    @classmethod
    def verify_install(cls, tools: ToolCache, **kwargs) -> Subprocess:
        """Make subprocess available in tool cache."""
        # short circuit since already verified and available
        if hasattr(tools, "subprocess"):
            return tools.subprocess

        tools.subprocess = Subprocess(tools=tools)
        return tools.subprocess

    @ensure_console_is_safe
    def run(
        self,
        args: SubprocessArgsT,
        stream_output: bool = True,
        filter_func: Callable[[str], Iterator[str]] | None = None,
        **kwargs,
    ) -> CompletedProcess:
        """A wrapper for subprocess.run().

        This method implements the behavior of subprocess.run() with some notable
        differences:

          - The command's output is streamed if ``stream_output=True`` or if dynamic
            console content such as the Wait Bar is active and I/O is not redirected.
          - If dynamic content is not active, ``stream_output=False`` will disable
            streaming.
          - If the ``env`` argument is provided, the current system environment will be
            copied, and the contents of env overwritten into that environment.
          - The ``text`` argument is defaulted to True so all output is returned as
            strings instead of bytes.

        Output Streaming Primer
        ~~~~~~~~~~~~~~~~~~~~~~~

        What is output streaming?
        -------------------------

        When a command is invoked via ``subprocess.run()``, by default, its stdout and
        stderr are connected directly to the console and the output is presented to the
        user as though the command was invoked natively. When output streaming is
        active, Briefcase reads the command's output and prints it in to the console on
        behalf of the command.

        Why is output streaming necessary?
        ----------------------------------

        Allowing commands to print directly to the console can provide superior UX since
        it allows tools to provide their native experience to users.

        However, a primary mechanism for providing support for Briefcase is asking users
        to provide the log file when errors occur. Often, these errors are raised by
        invoked commands; streaming the output of commands allows Briefcase to capture
        that output and include it in the log file as well as the console.

        Additionally, the content of the console can become corrupted if other processes
        print to the console while Rich's dynamic console elements such as a Progress
        Bar are active. Streaming such output allows Rich to maintain integrity of all
        the console content.

        How does output streaming work?
        -------------------------------

        Streaming is activated when:

          1. dynamic content such as the Wait Bar is active and I/O is not redirected

          2. the ``stream_output`` argument is ``True`` (as it is by default)

        I/O is considered redirected when both ``stdout`` and ``stderr`` are specified
        (e.g., when ``stdout`` and ``stderr`` are set to``subprocess.PIPE`` so the
        output is available to the caller).

        To facilitate streaming the output, ``stdout`` is set to ``subprocess.PIPE``. If
        ``stderr`` is not specified by calling code, then it is set to
        ``subprocess.STDOUT`` so it is interlaced with stdout output.

        The process is created for the command with ``Popen``, and a separate thread is
        started for continuously reading and printing the stdout to the console from the
        command process. It is necessary to read from stdout in a thread because the
        read is uninterruptible; so, a user may not be able to abort with CTRL-C.

        When the process ends (or the user sends CTRL-C), a call is made to ensure the
        command process terminates; this, in turn, ensures the thread for streaming can
        properly exit its otherwise infinite loop. The calling code is either returned a
        ``CompletedProcess`` object or ``CalledProcessError`` could be raised.

        Output streaming ultimately strives to simulate calling ``subprocess.run()``
        while proxying the command's output to the console.

        When *not* to use output streaming?
        -----------------------------------

        When streaming is disabled, the command's output **WILL NOT** be included the
        Briefcase log file. This makes troubleshooting difficult and should be avoided.
        However, there are some situations where disabling streaming is unavoidable.

        Some commands use their own dynamic console content such as progress bars,
        spinners, and other animations. When such output is streamed, its quality can be
        significantly compromised and look terrible in the console for users. If it is
        not possible to disable such output with e.g. a flag for the command, then
        disabling streaming may be necessary. As an example, the ``flatpak install``
        command uses many animations.

        If a command requires user input, do not use output streaming. The output
        streaming implementation doesn't provide any facilities to send input to the
        process and doesn't provide any special handling of stdin. Additionally, when
        the command process prompts the user for input, that doesn't always trigger
        ``readline()`` to return; so, the console may actually end up waiting for user
        input before displaying a prompt to the user. As an example, accepting Android
        SDK licenses requires user input.

        :param args: args for ``subprocess.run()``
        :param stream_output: simulate ``run()`` while streaming the output to the
            console. Set to False if the command requires user interaction.
        :param filter_func: If output streaming is enabled, a callable that will be
            invoked on every line of output that is streamed. The function accepts the
            "raw" line of input (stripped of any trailing newline); it returns a
            generator that yields the filtered output that should be displayed to the
            user. Can raise StopStreaming to terminate the output stream.
        :param kwargs: keyword args for ``subprocess.run()``
        :raises ValueError: if a filter function is provided when in non-streaming mode.
        :returns: ``CompletedProcess`` for invoked process
        """

        # Stream the output unless the caller explicitly disables it. When a
        # caller sets stream_output=False, then ensure_console_is_safe() will
        # disable any dynamic console elements while the command runs.
        if stream_output:
            return self._run_and_stream_output(args, filter_func=filter_func, **kwargs)

        # Can't filter non-streamed output.
        if filter_func:
            raise ValueError("Cannot apply a filter to non-streamed output.")

        # Otherwise, invoke run() normally.
        self._log_command(args)
        self._log_cwd(kwargs.get("cwd"))
        self._log_environment(kwargs.get("env"))

        try:
            command_result = self._subprocess.run(
                [str(arg) for arg in args], **self.final_kwargs(**kwargs)
            )
        except subprocess.CalledProcessError as e:
            self._log_return_code(e.returncode)
            raise
        self._log_return_code(command_result.returncode)

        return command_result

    def _run_and_stream_output(
        self,
        args: SubprocessArgsT,
        check: bool = False,
        filter_func: Callable[[str], Iterator[str]] | None = None,
        **kwargs,
    ) -> CompletedProcess:
        """Simulate subprocess.run() while streaming output to the console.

        This is useful when dynamic screen content is active or output should
        be captured for logging.

        When dynamic screen content like a Wait Bar is active, output can
        only be printed to the screen via Log or Console to avoid interfering
        with and likely corrupting the updates to the dynamic screen elements.

        stdout will always be piped so it can be printed to the screen;
        however, stderr can be piped by the caller so it is available
        to the caller in the return value...as with subprocess.run().

        Note: the 'timeout' and 'input' arguments are not supported.
        """
        if kwargs.get("stdout") and not kwargs.get("stderr"):
            # This is an unsupported configuration, as it's not clear where stderr
            # output would be displayed. Either:
            # * Redirect stderr in addition to stdout; or
            # * Use check_output() instead of run() to capture console output
            #   without streaming.
            raise AssertionError(
                "Subprocess.run() was invoked while dynamic Rich content is active (or via "
                "`stream_output`) with stdout redirected while stderr was not redirected."
            )
        for arg in [arg for arg in ["timeout", "input"] if arg in kwargs]:
            raise AssertionError(
                f"The Subprocess.run() '{arg}' argument is not supported "
                f"with `stream_output` or while dynamic Rich content is active."
            )

        # stdout must be piped so the output streamer can print it.
        kwargs["stdout"] = subprocess.PIPE
        # if stderr isn't explicitly redirected, then send it to stdout.
        kwargs.setdefault("stderr", subprocess.STDOUT)
        # use line-buffered output by default
        kwargs.setdefault("bufsize", 1)

        stderr = None
        with self.Popen(args, **kwargs) as process:
            self.stream_output(args[0], process, filter_func=filter_func)
            if process.stderr and kwargs["stderr"] != subprocess.STDOUT:
                stderr = process.stderr.read()
        return_code = process.poll()
        self._log_return_code(return_code)

        if check and return_code:
            raise subprocess.CalledProcessError(return_code, args, stderr=stderr)

        return subprocess.CompletedProcess(args, return_code, stderr=stderr)

    @ensure_console_is_safe
    def check_output(self, args: SubprocessArgsT, quiet: bool = False, **kwargs) -> str:
        """A wrapper for subprocess.check_output()

        The behavior of this method is identical to
        subprocess.check_output(), except for:
         - If the `env` is argument provided, the current system environment
           will be copied, and the contents of env overwritten into that
           environment.
         - The `text` argument is defaulted to True so all output
           is returned as strings instead of bytes.
         - The `stderr` argument is defaulted to `stdout` so _all_ output is
           returned and `stderr` isn't unexpectedly printed to the console.

        :param args: commands and its arguments to run via subprocess
        :param quiet: Should the invocation of this command be silent, and
            *not* appear in the logs? This should almost always be False;
            however, for some calls (most notably, calls that are called
            frequently to evaluate the status of another process), logging can
            be turned off so that log output isn't corrupted by thousands of
            polling calls.
        """
        # if stderr isn't explicitly redirected, then send it to stdout.
        kwargs.setdefault("stderr", subprocess.STDOUT)

        if not quiet:
            self._log_command(args)
            self._log_cwd(kwargs.get("cwd"))
            self._log_environment(kwargs.get("env"))

        try:
            cmd_output = self._subprocess.check_output(
                [str(arg) for arg in args], **self.final_kwargs(**kwargs)
            )
        except subprocess.CalledProcessError as e:
            if not quiet:
                self._log_output(e.output, e.stderr)
                self._log_return_code(e.returncode)
            raise

        if not quiet:
            self._log_output(cmd_output)
            self._log_return_code(0)
        return cmd_output

    def parse_output(
        self,
        output_parser: Callable[[str], ParserOutputT],
        args: SubprocessArgsT,
        **kwargs,
    ) -> ParserOutputT:
        """A wrapper for check_output() where the command output is processed through
        the supplied parser function.

        If the parser fails, CommandOutputParseError is raised. The parsing function
        should take one string argument and should raise ParseError for failure modes.

        :param output_parser: a function that takes str input and returns parsed
            content, or raises ParseError in the case of a parsing problem.
        :param args: The arguments to pass to the subprocess
        :param kwargs: The keyword arguments to pass to the subprocess
        :returns: Parsed data read from the subprocess output; the exact structure of
            that data is dependent on the output parser used.
        """
        cmd_output = self.check_output(args, **kwargs)

        try:
            return output_parser(cmd_output)
        except ParseError as e:
            error_reason = (
                str(e)
                or f"Failed to parse command output using '{output_parser.__name__}'"
            )
            self.tools.logger.error()
            self.tools.logger.error("Command Output Parsing Error:")
            self.tools.logger.error(f"    {error_reason}")
            self.tools.logger.error("Command:")
            self.tools.logger.error(
                f"    {' '.join(shlex.quote(str(arg)) for arg in args)}"
            )
            self.tools.logger.error("Command Output:")
            for line in ensure_str(cmd_output).splitlines():
                self.tools.logger.error(f"    {line}")
            raise CommandOutputParseError(error_reason) from e

    def Popen(self, args: SubprocessArgsT, **kwargs) -> subprocess.Popen:
        """A wrapper for subprocess.Popen()

        The behavior of this method is identical to
        subprocess.check_output(), except for:
         - If the `env` argument is provided, the current system environment
           will be copied, and the contents of env overwritten into that
           environment.
         - The `text` argument is defaulted to True so all output
           is returned as strings instead of bytes.
        """
        self._log_command(args)
        self._log_cwd(kwargs.get("cwd"))
        self._log_environment(kwargs.get("env"))

        return self._subprocess.Popen(
            [str(arg) for arg in args], **self.final_kwargs(**kwargs)
        )

    def stream_output(
        self,
        label: str,
        popen_process: subprocess.Popen,
        stop_func: Callable[[], bool] = lambda: False,
        filter_func: Callable[[str], Iterator[str]] | None = None,
    ):
        """Stream the output of a Popen process until the process exits. If the user
        sends CTRL+C, the process will be terminated.

        This is useful for starting a process via Popen such as tailing a log file, then
        initiating a non-blocking process that populates that log, and finally streaming
        the original process's output here.

        :param label: A description of the content being streamed; used for to provide
            context in logging messages.
        :param popen_process: a running Popen process with output to print
        :param stop_func: A Callable that returns True when output streaming should stop
            and the popen_process should be terminated.
        :param filter_func: A callable that will be invoked on every line of output that
            is streamed. The function accepts the "raw" line of input (stripped of any
            trailing newline); it returns a generator that yields the filtered output
            that should be displayed to the user. Can raise StopStreaming to terminate
            the output stream.
        """
        output_streamer = PopenOutputStreamer(
            label=label,
            popen_process=popen_process,
            logger=self.tools.logger,
            filter_func=filter_func,
        )
        try:
            output_streamer.start()
            # joining the thread is avoided due to demonstrated
            # instability of thread interruption via CTRL+C (#809)
            while not stop_func() and output_streamer.is_alive():
                time.sleep(0.1)
        except KeyboardInterrupt:
            self.tools.logger.info("Stopping...")
            # allow time for CTRL+C to propagate to the child process
            time.sleep(0.25)
            # re-raise to exit as "Aborted by user"
            raise
        finally:
            self.cleanup(label, popen_process)
            streamer_deadline = time.time() + 3
            while output_streamer.is_alive() and time.time() < streamer_deadline:
                time.sleep(0.1)
            if output_streamer.is_alive():
                self.tools.logger.error(
                    "Log stream hasn't terminated; log output may be corrupted."
                )

    def stream_output_non_blocking(
        self,
        label: str,
        popen_process: Popen,
        capture_output: bool = False,
        filter_func: Callable[[str], Iterator[str]] | None = None,
    ) -> PopenOutputStreamer:
        """Stream the output of a Popen process without blocking.

        This is useful for streaming or capturing the output of a process in the
        background. In this way, the process' output can be shown to users while the
        main thread monitors other activities; alternatively, the output of the process
        can be captured to be retrieved later in the event of an error, for instance.

        :param label: A description of the content being streamed; used for to provide
            context in logging messages.
        :param popen_process: A running Popen process with output to print
        :param capture_output: Retain process output instead of printing to the console
        :param filter_func: A callable that will be invoked on every line of output that
            is streamed. The function accepts the "raw" line of input (stripped of any
            trailing newline); it returns a generator that yields the filtered output
            that should be displayed to the user. Can raise StopStreaming to terminate
            the output stream.
        """
        output_streamer = PopenOutputStreamer(
            label=label,
            popen_process=popen_process,
            logger=self.tools.logger,
            capture_output=capture_output,
            filter_func=filter_func,
        )
        output_streamer.start()
        return output_streamer

    def cleanup(self, label: str, popen_process: subprocess.Popen):
        """Clean up after a Popen process, gracefully terminating if possible; forcibly
        if not.

        :param label: A description of the content being streamed; used for to provide
            context in logging messages.
        :param popen_process: The Popen instance to clean up.
        """
        popen_process.terminate()
        try:
            popen_process.wait(timeout=3)
        except subprocess.TimeoutExpired:
            self.tools.logger.warning(f"Forcibly killing {label}...")
            popen_process.kill()

    def _log(self, msg: str = ""):
        """Funnel for all subprocess details logging."""
        self.tools.logger.debug(msg, preface=">>> " if msg else "")

    def _log_command(self, args: SubprocessArgsT):
        """Log the entire console command being executed."""
        self._log()
        self._log("Running Command:")
        self._log(f"    {' '.join(shlex.quote(str(arg)) for arg in args)}")

    def _log_cwd(self, cwd: str | Path | None):
        """Log the working directory for the command being executed."""
        effective_cwd = Path.cwd() if cwd is None else cwd
        self._log("Working Directory:")
        self._log(f"    {effective_cwd}")

    def _log_environment(self, overrides: dict[str, str] | None):
        """Log the environment variables overrides prior to command execution.

        :param overrides: The explicit environment passed to the subprocess call;
            can be `None` if there are no explicit environment changes.
        """
        if overrides:
            self._log("Environment Overrides:")
            for env_var, value in overrides.items():
                self._log(f"    {env_var}={value}")

    def _log_output(self, output: str, stderr: str | None = None):
        """Log the output of the executed command."""
        if output:
            self._log("Command Output:")
            for line in ensure_str(output).splitlines():
                self._log(f"    {line}")

        if stderr:
            self._log("Command Error Output (stderr):")
            for line in ensure_str(stderr).splitlines():
                self._log(f"    {line}")

    def _log_return_code(self, return_code: int | str):
        """Log the output value of the executed command."""
        self._log(f"Return code: {return_code}")



import os
from unittest import mock

import pytest
from cookiecutter.main import cookiecutter

import briefcase
from briefcase.commands import ConvertCommand
from briefcase.console import Console, Log


@pytest.fixture
def convert_command(tmp_path):
    return ConvertCommand(
        base_path=tmp_path / "project",
        data_path=tmp_path / "data",
        logger=Log(),
        console=Console(),
    )


def test_convert_app_unused_project_overrides(
    monkeypatch,
    convert_command,
    tmp_path,
    capsys,
):
    """The user is informed of unused project configuration overrides."""
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    app_context = {
        "formal_name": "My Application",
        "class_name": "MyApplication",
        "app_name": "myapplication",
        "module_name": "mymodule",
        "test_source_dir": "test_files",
    }
    convert_command.build_app_context = mock.MagicMock(return_value=app_context)
    convert_command.build_gui_context = mock.MagicMock(
        return_value={"gui_framework": "None"}
    )
    convert_command.update_cookiecutter_cache = mock.MagicMock(
        return_value="~/.cookiecutters/briefcase-template"
    )
    convert_command.tools.cookiecutter = mock.MagicMock(spec_set=cookiecutter)
    convert_command.migrate_necessary_files = mock.MagicMock()

    # Create the new app, using the default template.
    convert_command.convert_app(
        tmp_path=tmp_path / "working",
        project_overrides={"unused": "override"},
    )

    # App context is constructed
    convert_command.build_app_context.assert_called_once_with({"unused": "override"})
    convert_command.build_gui_context.assert_called_once_with(
        app_context, {"unused": "override"}
    )
    # Template is updated
    convert_command.update_cookiecutter_cache.assert_called_once_with(
        template="https://github.com/beeware/briefcase-template",
        branch="v37.42.7",
    )
    # Cookiecutter is invoked
    convert_command.tools.cookiecutter.assert_called_once_with(
        "~/.cookiecutters/briefcase-template",
        no_input=True,
        output_dir=os.fsdecode(tmp_path / "working"),
        checkout="v37.42.7",
        extra_context={
            "formal_name": "My Application",
            "class_name": "MyApplication",
            "app_name": "myapplication",
            "module_name": "mymodule",
            "test_source_dir": "test_files",
            # The expected app context should now also contain the default template,
            # branch, and Briefcase version.
            "template_source": "https://github.com/beeware/briefcase-template",
            "template_branch": "v37.42.7",
            "briefcase_version": "37.42.7",
            "gui_framework": "None",
        },
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )
    convert_command.migrate_necessary_files.assert_called_once_with(
        tmp_path / "working" / app_context["app_name"],
        app_context["test_source_dir"],
        "mymodule",
    )

    unused_project_override_warning = (
        "WARNING: These project configuration overrides were not used:\n\n"
        "    unused = override"
    )
    assert unused_project_override_warning in capsys.readouterr().out



from unittest.mock import MagicMock

from ...utils import PartialMatchString


def test_app_name_is_formatted(convert_command, monkeypatch):
    mock_input_text = MagicMock()
    monkeypatch.setattr(convert_command, "input_text", mock_input_text)
    convert_command.input_formal_name("test-app-name", None)

    mock_input_text.assert_called_once_with(
        intro=PartialMatchString(
            "Based on the app name, we suggest a formal name of 'Test App Name',"
        ),
        variable="formal name",
        default="Test App Name",
        override_value=None,
    )


def test_override_is_used(convert_command):
    assert (
        convert_command.input_formal_name("test-app-name", "OVERRIDE_VALUE")
        == "OVERRIDE_VALUE"
    )


def test_prompted_formal_name(convert_command):
    """You can type in the formal name."""
    convert_command.input.values = ["My Formal Name"]
    assert convert_command.input_formal_name("app-name", None) == "My Formal Name"



import pytest

from briefcase.exceptions import MissingToolError, NetworkFailure


def test_upgrade_exists(mock_tools, rcedit, tmp_path):
    """If rcedit already exists, upgrading deletes first."""
    rcedit_path = tmp_path / "tools/rcedit-x64.exe"

    # Mock the existence of an install
    rcedit_path.touch()

    # Mock a successful download
    def side_effect_create_mock_appimage(*args, **kwargs):
        rcedit_path.touch()
        return "new-downloaded-file"

    mock_tools.file.download.side_effect = side_effect_create_mock_appimage

    # Do upgrade
    rcedit.upgrade()

    # The mock file should exist as the upgraded version
    assert rcedit_path.exists()

    # A download is invoked
    mock_tools.file.download.assert_called_with(
        url="https://github.com/electron/rcedit/"
        "releases/download/v2.0.0/rcedit-x64.exe",
        download_path=tmp_path / "tools",
        role="RCEdit",
    )


def test_upgrade_does_not_exist(mock_tools, rcedit, tmp_path):
    """If rcedit doesn't already exist, upgrading is an error."""
    # Do upgrade
    with pytest.raises(MissingToolError):
        rcedit.upgrade()

    # The tool wasn't already installed, so an error is raised.
    assert mock_tools.file.download.call_count == 0


def test_upgrade_rcedit_download_failure(mock_tools, rcedit, tmp_path):
    """If rcedit doesn't exist, but a download failure occurs, an error is raised."""
    # Mock the existence of an install
    rcedit_path = tmp_path / "tools/rcedit-x64.exe"
    rcedit_path.touch()

    mock_tools.file.download.side_effect = NetworkFailure("mock")

    # The upgrade will fail
    with pytest.raises(NetworkFailure, match="Unable to mock"):
        rcedit.upgrade()

    # The mock file will be deleted
    assert not rcedit_path.exists()

    # A download was invoked
    mock_tools.file.download.assert_called_with(
        url="https://github.com/electron/rcedit/"
        "releases/download/v2.0.0/rcedit-x64.exe",
        download_path=tmp_path / "tools",
        role="RCEdit",
    )



from unittest.mock import MagicMock

import pytest

from briefcase.integrations.base import ToolCache
from briefcase.integrations.file import File
from briefcase.integrations.subprocess import Subprocess

JDK_RELEASE = "17.0.11"
JDK_BUILD = "9"


@pytest.fixture
def mock_tools(mock_tools, tmp_path) -> ToolCache:
    # Mock default tools
    mock_tools.subprocess = MagicMock(spec_set=Subprocess)
    mock_tools.file.download = MagicMock(spec_set=File.download)

    return mock_tools



import pytest

from briefcase.commands.base import BaseCommand
from briefcase.config import AppConfig
from briefcase.console import Console, Log


class DummyCommand(BaseCommand):
    """A dummy command to test the BaseCommand interface."""

    command = ("dummy",)
    # Platform and format contain upper case to test case normalization
    platform = "Tester"
    output_format = "Dummy"
    description = "Dummy base command"

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("logger", Log())
        kwargs.setdefault("console", Console())
        super().__init__(*args, **kwargs)

        self.actions = []

    def add_options(self, parser):
        # Provide some extra arguments:
        # * some optional arguments
        parser.add_argument("-x", "--extra")
        parser.add_argument("-m", "--mystery")
        # * a required argument
        parser.add_argument("-r", "--required", required=True)

    def binary_path(self, app):
        raise NotImplementedError()

    def verify_host(self):
        super().verify_host()
        self.actions.append(("verify-host",))

    def verify_tools(self):
        super().verify_tools()
        self.actions.append(("verify-tools",))

    def finalize_app_config(self, app):
        super().finalize_app_config(app=app)
        self.actions.append(("finalize-app-config", app.app_name))


@pytest.fixture
def base_command(tmp_path):
    command = DummyCommand(
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "data_path",
    )
    command.parse_options(["-r", "default"])
    return command


# Define some stub command classes
# These will be used to test the command accessor
class DummyCreateCommand(DummyCommand):
    description = "Test Create"


class DummyUpdateCommand(DummyCommand):
    description = "Test Update"


class DummyBuildCommand(DummyCommand):
    description = "Test Build"


class DummyRunCommand(DummyCommand):
    description = "Test Run"


class DummyPackageCommand(DummyCommand):
    description = "Test Package"


class DummyPublishCommand(DummyCommand):
    description = "Test Publish"


# Register the commands with the module
create = DummyCreateCommand
update = DummyUpdateCommand
build = DummyBuildCommand
run = DummyRunCommand
package = DummyPackageCommand
publish = DummyPublishCommand


class OtherDummyCommand(BaseCommand):
    command = ("other",)
    platform = "tester"
    output_format = "dumdum"
    description = "Another dummy command"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def binary_path(self, app):
        raise NotImplementedError()


@pytest.fixture
def other_command(tmp_path):
    return OtherDummyCommand(base_path=tmp_path, logger=Log(), console=Console())


@pytest.fixture
def my_app():
    return AppConfig(
        app_name="my-app",
        formal_name="My App",
        bundle="com.example",
        version="1.2.3",
        description="This is a simple app",
        sources=["src/my_app"],
        license={"file": "LICENSE"},
    )



import os
import platform
import subprocess
from contextlib import suppress
from pathlib import Path

import pytest

from briefcase.exceptions import BriefcaseCommandError

from .conftest import DummyCommand


@pytest.mark.skipif(platform.system() != "Windows", reason="Windows specific tests")
def test_path_is_realpath(tmp_path):
    """Briefcase's data path matches its realpath.

    For the Windows Store Python, filesystem interaction with `%LOCALAPPDATA%` can be
    redirected to a sandboxed location. However, `os.path.realpath()` will reveal such a
    redirection. This test ensures that Briefcase bypasses the sandboxing.
    """
    data_path = Path(os.environ["LOCALAPPDATA"]) / "realpathbase"
    try:
        command = DummyCommand(data_path=data_path)
        assert command.data_path == Path(os.path.realpath(data_path))
    finally:
        with suppress(FileNotFoundError):
            os.rmdir(data_path)


def test_data_path_creation_failure(tmp_path, monkeypatch):
    """An error is raised if the data path cannot be created."""

    def raise_calledprpcesserror(*a, **kw):
        raise subprocess.CalledProcessError(returncode=1, cmd=["mkdir", str(data_path)])

    # On Linux and macOS, use /etc/ to raise an OSError
    data_path = Path("/etc/mydatadir")

    # Patch run() since it's apparently quite difficult to find a
    # location in Windows that is guaranteed to throw an error...
    monkeypatch.setattr(subprocess, "run", raise_calledprpcesserror)

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Failed to create the Briefcase directory to store tools and support files:",
    ):
        DummyCommand(data_path=data_path)


def test_space_in_path(tmp_path):
    """The briefcase data path cannot contain spaces."""
    with pytest.raises(
        BriefcaseCommandError,
        match=r"contains spaces. This will cause problems with some tools",
    ):
        DummyCommand(base_path=tmp_path / "base", data_path=tmp_path / "somewhere bad")


def test_empty_custom_path(monkeypatch, tmp_path):
    """If the environment-specified BRIEFCASE_HOME is defined, but empty, an error is
    raised."""
    monkeypatch.setenv("BRIEFCASE_HOME", "")

    with pytest.raises(
        BriefcaseCommandError,
        match=r"The path specified by BRIEFCASE_HOME does not exist.",
    ):
        DummyCommand(base_path=tmp_path / "base")


def test_custom_path_does_not_exist(monkeypatch, tmp_path):
    """If the environment-specified BRIEFCASE_HOME doesn't exist, an error is raised."""
    monkeypatch.setenv("BRIEFCASE_HOME", str(tmp_path / "custom"))

    with pytest.raises(
        BriefcaseCommandError,
        match=r"The path specified by BRIEFCASE_HOME does not exist.",
    ):
        DummyCommand(base_path=tmp_path / "base")


def templated_path_test(
    monkeypatch,
    tmp_path,
    data_path,
    environ_path,
    expected_data_path,
):
    if environ_path:
        monkeypatch.setenv("BRIEFCASE_HOME", environ_path.format(tmp_path=tmp_path))
        Path(environ_path.format(tmp_path=tmp_path)).mkdir(parents=True)
    else:
        monkeypatch.delenv("BRIEFCASE_HOME", raising=False)

    command = DummyCommand(
        base_path=tmp_path / "base",
        data_path=data_path.format(tmp_path=tmp_path) if data_path else None,
    )

    assert command.base_path == tmp_path / "base"
    assert command.data_path.exists()
    assert command.data_path == Path(
        os.path.expanduser(expected_data_path.format(tmp_path=tmp_path))
    )


@pytest.mark.skipif(platform.system() != "Darwin", reason="macOS specific tests")
@pytest.mark.parametrize(
    "data_path, environ_path, expected_data_path",
    [
        (  # All default values
            None,
            None,
            "~/Library/Caches/org.beeware.briefcase",
        ),
        (  # Environment variable for the data home
            None,
            "{tmp_path}custom/briefcase/path",
            "{tmp_path}custom/briefcase/path",
        ),
        (  # Explicit paths
            "{tmp_path}path/to/data",
            None,
            "{tmp_path}path/to/data",
        ),
        (  # Explicit paths and an environment variable present
            "{tmp_path}path/to/data",
            "{tmp_path}custom/briefcase/path",
            "{tmp_path}path/to/data",
        ),
    ],
)
def test_macOS_paths(
    monkeypatch,
    tmp_path,
    data_path,
    environ_path,
    expected_data_path,
):
    templated_path_test(
        monkeypatch,
        tmp_path,
        data_path,
        environ_path,
        expected_data_path,
    )


@pytest.mark.skipif(platform.system() != "Windows", reason="Windows specific tests")
@pytest.mark.parametrize(
    "data_path, environ_path, expected_data_path",
    [
        (  # All default values
            None,
            None,
            "~/AppData/Local/BeeWare/briefcase/Cache",
        ),
        (  # Environment variable for the data home
            None,
            "{tmp_path}custom\\briefcase\\path",
            "{tmp_path}custom\\briefcase\\path",
        ),
        (  # Explicit paths
            "{tmp_path}path\\to\\data",
            None,
            "{tmp_path}path\\to\\data",
        ),
        (  # Explicit paths and an environment variable present
            "{tmp_path}path\\to\\data",
            "{tmp_path}\\briefcase\\path",
            "{tmp_path}path\\to\\data",
        ),
    ],
)
def test_windows_paths(
    monkeypatch,
    tmp_path,
    data_path,
    environ_path,
    expected_data_path,
):
    templated_path_test(
        monkeypatch,
        tmp_path,
        data_path,
        environ_path,
        expected_data_path,
    )


@pytest.mark.skipif(platform.system() != "Linux", reason="Linux specific tests")
@pytest.mark.parametrize(
    "data_path, environ_path, expected_data_path",
    [
        (  # All default values
            None,
            None,
            "~/.cache/briefcase",
        ),
        (  # Environment variable for the data home
            None,
            "{tmp_path}custom/briefcase/path",
            "{tmp_path}custom/briefcase/path",
        ),
        (  # Explicit paths
            "{tmp_path}path/to/data",
            None,
            "{tmp_path}path/to/data",
        ),
        (  # Explicit paths and an environment variable present
            "{tmp_path}path/to/data",
            "{tmp_path}custom/briefcase/path",
            "{tmp_path}path/to/data",
        ),
    ],
)
def test_linux_paths(
    monkeypatch,
    tmp_path,
    data_path,
    environ_path,
    expected_data_path,
):
    templated_path_test(
        monkeypatch,
        tmp_path,
        data_path,
        environ_path,
        expected_data_path,
    )



import pytest

from briefcase.config import is_reserved_keyword


@pytest.mark.parametrize(
    "name",
    [
        # These names may not be valid identifiers,
        # but they're not reserved words in any language we care about.
        "BLACK",
        "purple",
        "9pink",
        "Yellow",
        "green",
        "hello world",
    ],
)
def test_is_not_reserved_keyword_violation(name):
    """Test that names not in the keywords list return false."""
    assert not is_reserved_keyword(name)


@pytest.mark.parametrize(
    "name",
    [
        # Python specific keyword, in various case variants
        "pass",
        "Pass",
        "PASS",
        # Javascript specific keyword, in various case variants
        "in",
        "In",
        "IN",
        # Java specific keyword, in various case variants
        "synchronized",
        "Synchronized",
        "SYNCHRONIZED",
        # Keyword in multiple languages, in various case variants
        "false",
        "False",
        "FALSE",
        # Windows reserved keywords
        "CON",
        "con",  # lower case version of reserved name
        "LPT5",
        "Lpt5",  # unconventional spelling of reserved name
    ],
)
def test_is_reserved_keyword(name):
    """Test that names in the reserved_keywords list returns true."""
    assert is_reserved_keyword(name)



import os
from pathlib import Path
from unittest import mock

import pytest

from briefcase.exceptions import BriefcaseCommandError


def test_missing_avd_config(android_sdk):
    """If an AVD config can't be found, raise an error."""
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to read configuration of AVD @unknownDevice",
    ):
        android_sdk.verify_avd("unknownDevice")


def test_minimal_config(android_sdk, capsys):
    """If the AVD config doesn't contain any interesting keys, raise a warning, but
    continue."""
    # Mock an AVD configuration that doesn't contain an image.sysdir.1,
    # skin.name or skin.path key.
    android_sdk.avd_config = mock.MagicMock(
        return_value={
            "avd.ini.encoding": "UTF-8",
            "hw.device.manufacturer": "Google",
            "hw.device.name": "pixel",
            "weird.key": "good=bad",
            "PlayStore.enabled": "no",
            "avd.name": "beePhone",
            "disk.cachePartition": "yes",
            "disk.cachePartition.size": "42M",
        }
    )
    android_sdk.verify_system_image = mock.MagicMock()
    android_sdk.verify_emulator_skin = mock.MagicMock()

    # Verify the AVD
    android_sdk.verify_avd("minimalDevice")

    # The AVD config was loaded.
    android_sdk.avd_config.assert_called_with("minimalDevice")

    # A warning message was output
    assert "WARNING: Unable to determine AVD system image" in capsys.readouterr().out

    # The system image was not verified
    android_sdk.verify_system_image.assert_not_called()

    # An emulator skin was not verified
    android_sdk.verify_emulator_skin.assert_not_called()


def test_valid_system_image(android_sdk):
    """If the AVD config contains an image.sysdir.1 key, it is used to create a system
    image name."""
    # Mock an AVD configuration that contains an image.sysdir.1 key
    android_sdk.avd_config = mock.MagicMock(
        return_value={
            "avd.ini.encoding": "UTF-8",
            "hw.device.manufacturer": "Google",
            "hw.device.name": "pixel",
            "weird.key": "good=bad",
            "PlayStore.enabled": "no",
            "avd.name": "beePhone",
            "disk.cachePartition": "yes",
            "disk.cachePartition.size": "42M",
            # Add an OS-dependent image.sysdir.1 value, with a trailing slash.
            "image.sysdir.1": os.fsdecode(
                Path("system-images/android-31/default/arm64-v8a")
            )
            + "/",
        }
    )
    android_sdk.verify_system_image = mock.MagicMock()
    android_sdk.verify_emulator_skin = mock.MagicMock()

    # Verify the AVD
    android_sdk.verify_avd("goodDevice")

    # The AVD config was loaded.
    android_sdk.avd_config.assert_called_with("goodDevice")

    # The system image was not verified
    android_sdk.verify_system_image.assert_called_once_with(
        "system-images;android-31;default;arm64-v8a"
    )

    # The emulator skin was not verified
    android_sdk.verify_emulator_skin.assert_not_called()


def test_valid_emulator_skin(android_sdk):
    """If the AVD config contains a known emulator skin type, it is verified."""
    # Mock an AVD configuration that contains skin.name and skin.path keys
    android_sdk.avd_config = mock.MagicMock(
        return_value={
            "avd.ini.encoding": "UTF-8",
            "hw.device.manufacturer": "Google",
            "hw.device.name": "pixel",
            "weird.key": "good=bad",
            "PlayStore.enabled": "no",
            "avd.name": "beePhone",
            "disk.cachePartition": "yes",
            "disk.cachePartition.size": "42M",
            # Add an emulator skin.
            "skin.name": "pixel_3a",
            "skin.path": "skins/pixel_3a",
        }
    )
    android_sdk.verify_system_image = mock.MagicMock()
    android_sdk.verify_emulator_skin = mock.MagicMock()

    # Verify the AVD
    android_sdk.verify_avd("goodDevice")

    # The AVD config was loaded.
    android_sdk.avd_config.assert_called_with("goodDevice")

    # The system image was not verified
    android_sdk.verify_system_image.assert_not_called()

    # The emulator skin will be verified
    android_sdk.verify_emulator_skin.assert_called_with("pixel_3a")


def test_unrecognized_emulator_skin(android_sdk, capsys):
    """If the AVD config contains an emulator skin in an unusual location, raise a
    warning, but continue."""
    # Mock an AVD configuration that contains a skin.name and skin.path
    # in an unexpected location
    android_sdk.avd_config = mock.MagicMock(
        return_value={
            "avd.ini.encoding": "UTF-8",
            "hw.device.manufacturer": "Google",
            "hw.device.name": "pixel",
            "weird.key": "good=bad",
            "PlayStore.enabled": "no",
            "avd.name": "beePhone",
            "disk.cachePartition": "yes",
            "disk.cachePartition.size": "42M",
            # Add an emulator skin.
            "skin.name": "pixel_3a",
            "skin.path": "weird/pixel_3a",
        }
    )
    android_sdk.verify_system_image = mock.MagicMock()
    android_sdk.verify_emulator_skin = mock.MagicMock()

    # Verify the AVD
    android_sdk.verify_avd("goodDevice")

    # The AVD config was loaded.
    android_sdk.avd_config.assert_called_with("goodDevice")

    # A warning message was output
    assert "WARNING: Unrecognized device skin" in capsys.readouterr().out

    # The system image was not verified
    android_sdk.verify_system_image.assert_not_called()

    # The emulator skin was not verified
    android_sdk.verify_emulator_skin.assert_not_called()


def test_no_skin_emulator_skin(android_sdk, capsys):
    """If the AVD config contains an emulator skin set to _no_skin continue without
    validation."""
    # Mock an AVD configuration that contains a skin.name and skin.path
    # in an unexpected location
    android_sdk.avd_config = mock.MagicMock(
        return_value={
            "avd.ini.encoding": "UTF-8",
            "hw.device.manufacturer": "Google",
            "hw.device.name": "pixel",
            "weird.key": "good=bad",
            "PlayStore.enabled": "no",
            "avd.name": "beePhone",
            "disk.cachePartition": "yes",
            "disk.cachePartition.size": "42M",
            # Add an emulator skin.
            "skin.name": "768x1280",
            "skin.path": "_no_skin",
        }
    )
    android_sdk.verify_system_image = mock.MagicMock()
    android_sdk.verify_emulator_skin = mock.MagicMock()

    # Verify the AVD
    android_sdk.verify_avd("goodDevice")

    # The AVD config was loaded.
    android_sdk.avd_config.assert_called_with("goodDevice")

    # A warning message should not be output
    assert "WARNING: Unrecognized device skin" not in capsys.readouterr().out

    # The system image was not verified
    android_sdk.verify_system_image.assert_not_called()

    # The emulator skin was not verified
    android_sdk.verify_emulator_skin.assert_not_called()



from unittest import mock

import pytest

from briefcase.console import Console, Log
from briefcase.integrations.flatpak import Flatpak
from briefcase.platforms.linux.flatpak import LinuxFlatpakPackageCommand


@pytest.fixture
def package_command(tmp_path):
    return LinuxFlatpakPackageCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


def test_package(package_command, first_app_config, tmp_path):
    """A flatpak can be packaged."""
    package_command.tools.flatpak = mock.MagicMock(spec_set=Flatpak)
    # Mock the architecture so that the output filename is consistent
    package_command.tools.host_arch = "gothic"

    first_app_config.flatpak_runtime_repo_url = "https://example.com/flatpak/repo"
    first_app_config.flatpak_runtime_repo_alias = "custom-repo"

    package_command.package_app(first_app_config)

    # A flatpak bundle is created
    package_command.tools.flatpak.bundle.assert_called_once_with(
        repo_url="https://example.com/flatpak/repo",
        bundle_identifier="com.example.first-app",
        app_name="first-app",
        version="0.0.1",
        build_path=tmp_path / "base_path/build/first-app/linux/flatpak",
        output_path=tmp_path / "base_path/dist/First_App-0.0.1-gothic.flatpak",
    )
